# sage.doctest: needs sage.libs.pari
r"""
Dimensions of spaces of modular forms

AUTHORS:

- William Stein

- Jordi Quer

ACKNOWLEDGEMENT: The dimension formulas and implementations in this
module grew out of a program that Bruce Kaskel wrote (around 1996)
in PARI, which Kevin Buzzard subsequently extended. I (William
Stein) then implemented it in C++ for Hecke. I also implemented it
in Magma. Also, the functions for dimensions of spaces with
nontrivial character are based on a paper (that has no proofs) by
Cohen and Oesterlé [CO1977]_. The formulas for `\Gamma_H(N)` were found
and implemented by Jordi Quer.

The formulas here are more complete than in Hecke or Magma.

Currently the input to each function below is an integer and either a Dirichlet
character `\varepsilon` or a finite index subgroup of `\SL_2(\ZZ)`.
If the input is a Dirichlet character `\varepsilon`, the dimensions are for
subspaces of `M_k(\Gamma_1(N), \varepsilon)`, where `N` is the modulus of
`\varepsilon`.

These functions mostly call the methods ``dimension_cusp_forms``,
``dimension_modular_forms`` and so on of the corresponding congruence subgroup
classes.

REFERENCES:

.. [CO1977] \H. Cohen, J. Oesterlé, *Dimensions des espaces de formes
   modulaires*, p. 69-78 in Modular functions in one variable VI.
   Lecture Notes in Math. 627, Springer-Verlag, NewYork, 1977.
"""
# ****************************************************************************
#       Copyright (C) 2004-2008 William Stein <wstein@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.arith.misc import factor, is_prime, valuation
from sage.misc.misc_c import prod
from sage.modular.arithgroup.congroup_gamma0 import Gamma0_constructor as Gamma0
from sage.modular.arithgroup.congroup_gamma1 import Gamma1_constructor as Gamma1
from sage.modular.arithgroup.congroup_gammaH import GammaH_class
from sage.modular.arithgroup.congroup_generic import ArithmeticSubgroup
from sage.rings.finite_rings.integer_mod import Mod
from sage.rings.finite_rings.integer_mod_ring import IntegerModRing
from sage.rings.integer import Integer
from sage.rings.rational_field import frac

from sage.modular import dirichlet

##########################################################################
# Helper functions for calculating dimensions of spaces of modular forms
##########################################################################


def eisen(p):
    """
    Return the Eisenstein number `n` which is the numerator of `(p-1)/12`.

    INPUT:

    - ``p`` -- a prime

    OUTPUT: integer

    EXAMPLES::

        sage: [(p, sage.modular.dims.eisen(p)) for p in prime_range(24)]
        [(2, 1), (3, 1), (5, 1), (7, 1), (11, 5), (13, 1), (17, 4),
         (19, 3), (23, 11)]
    """
    if not is_prime(p):
        raise ValueError("p must be prime")
    return frac(p - 1, 12).numerator()

##########################################################################
# Formula of Cohen-Oesterlé for dim S_k(Gamma_1(N),eps).  REF:
# Springer Lecture notes in math, volume 627, pages 69--78.  The
# functions CO_delta and CO_nu, which were first written by Kevin
# Buzzard, are used only by the function CohenOesterle.
##########################################################################


def CO_delta(r, p, N, eps):
    r"""
    This is used as an intermediate value in computations related to
    the paper of Cohen-Oesterlé.

    INPUT:

    - ``r`` -- positive integer

    - ``p`` -- a prime

    - ``N`` -- positive integer

    - ``eps`` -- character

    OUTPUT: element of the base ring of the character

    EXAMPLES::

        sage: G.<eps> = DirichletGroup(7)
        sage: sage.modular.dims.CO_delta(1,5,7,eps^3)
        2
    """
    K = eps.base_ring()
    if p % 4 == 3:
        return K.zero()
    if p == 2:
        if r == 1:
            return K.one()
        return K.zero()
    # interesting case: p=1(mod 4).
    # omega is a primitive 4th root of unity mod p.
    omega = (IntegerModRing(p).unit_gens()[0])**((p - 1) // 4)
    # this n is within a p-power root of a "local" 4th root of 1 modulo p.
    n = Mod(int(omega.crt(Mod(1, N // (p**r)))), N)
    n = n**(p**(r - 1))   # this is correct now
    t = eps(n)
    if t == K.one():
        return K(2)
    if t == -K.one():
        return K(-2)
    return K.zero()


def CO_nu(r, p, N, eps):
    r"""
    This is used as an intermediate value in computations related to
    the paper of Cohen-Oesterlé.

    INPUT:

    - ``r`` -- positive integer

    - ``p`` -- a prime

    - ``N`` -- positive integer

    - ``eps`` -- character

    OUTPUT: element of the base ring of the character

    EXAMPLES::

        sage: G.<eps> = DirichletGroup(7)
        sage: G.<eps> = DirichletGroup(7)
        sage: sage.modular.dims.CO_nu(1,7,7,eps)
        -1
    """
    K = eps.base_ring()
    if p % 3 == 2:
        return K.zero()
    if p == 3:
        if r == 1:
            return K.one()
        return K.zero()
    # interesting case: p=1(mod 3)
    # omega is a cube root of 1 mod p.
    omega = (IntegerModRing(p).unit_gens()[0])**((p - 1) // 3)
    n = Mod(omega.crt(Mod(1, N // (p**r))), N)
    # within a p-power root of a "local" cube root of 1 mod p.
    n = n**(p**(r - 1))  # this is right now
    t = eps(n)
    if t == K.one():
        return K(2)
    return -K.one()


def CohenOesterle(eps, k):
    r"""
    Compute the Cohen-Oesterlé function associate to eps, `k`.

    This is a summand in the formula for the dimension of the space of
    cusp forms of weight `2` with character `\varepsilon`.

    INPUT:

    - ``eps`` -- Dirichlet character

    - ``k`` -- integer

    OUTPUT: element of the base ring of eps

    EXAMPLES::

        sage: G.<eps> = DirichletGroup(7)
        sage: sage.modular.dims.CohenOesterle(eps, 2)
        -2/3
        sage: sage.modular.dims.CohenOesterle(eps, 4)
        -1
    """
    N = eps.modulus()
    facN = factor(N)
    f = eps.conductor()
    gamma_k = 0
    if k % 4 == 2:
        gamma_k = frac(-1, 4)
    elif k % 4 == 0:
        gamma_k = frac(1, 4)
    mu_k = 0
    if k % 3 == 2:
        mu_k = frac(-1, 3)
    elif k % 3 == 0:
        mu_k = frac(1, 3)

    def _lambda(r, s, p):
        """
        Used internally by the CohenOesterle function.

        INPUT:

        - ``r``, ``s``, ``p`` -- integers

        OUTPUT: integer

        EXAMPLES: (indirect doctest)

        ::

            sage: # needs sage.rings.number_field
            sage: K = CyclotomicField(3)
            sage: eps = DirichletGroup(7*43, K).0^2
            sage: sage.modular.dims.CohenOesterle(eps, 2)
            -4/3
        """
        if 2 * s <= r:
            if r % 2 == 0:
                return p**(r // 2) + p**((r // 2) - 1)
            return 2 * p**((r - 1) // 2)
        return 2 * p**(r - s)

    K = eps.base_ring()
    return K(frac(-1, 2) *
             prod(_lambda(r, valuation(f, p), p) for p, r in facN) +
             gamma_k * K.prod(CO_delta(r, p, N, eps) for p, r in facN) +
             mu_k * K.prod(CO_nu(r, p, N, eps) for p, r in facN))


####################################################################
# Functions exported to the global namespace.
# These have very flexible inputs.
####################################################################

def dimension_new_cusp_forms(X, k=2, p=0):
    """
    Return the dimension of the new (or `p`-new) subspace of
    cusp forms for the character or group `X`.

    INPUT:

    - ``X`` -- integer, congruence subgroup or Dirichlet
      character

    - ``k`` -- weight (integer)

    - ``p`` -- 0 or a prime

    EXAMPLES::

        sage: from sage.modular.dims import dimension_new_cusp_forms
        sage: dimension_new_cusp_forms(100,2)
        1

        sage: dimension_new_cusp_forms(Gamma0(100),2)
        1
        sage: dimension_new_cusp_forms(Gamma0(100),4)
        5

        sage: dimension_new_cusp_forms(Gamma1(100),2)
        141
        sage: dimension_new_cusp_forms(Gamma1(100),4)
        463

        sage: dimension_new_cusp_forms(DirichletGroup(100).1^2,2)
        2
        sage: dimension_new_cusp_forms(DirichletGroup(100).1^2,4)
        8

        sage: sum(dimension_new_cusp_forms(e,3) for e in DirichletGroup(30))
        12
        sage: dimension_new_cusp_forms(Gamma1(30),3)
        12

    Check that :issue:`12640` is fixed::

        sage: dimension_new_cusp_forms(DirichletGroup(1)(1), 12)
        1
        sage: dimension_new_cusp_forms(DirichletGroup(2)(1), 24)
        1
    """
    if isinstance(X, GammaH_class):
        return X.dimension_new_cusp_forms(k, p=p)
    if isinstance(X, dirichlet.DirichletCharacter):
        N = X.modulus()
        if N <= 2:
            return Gamma0(N).dimension_new_cusp_forms(k, p=p)
        # Gamma1(N) for N<=2 just returns Gamma0(N), which has no
        # eps parameter. See trac #12640.
        return Gamma1(N).dimension_new_cusp_forms(k, eps=X, p=p)
    if isinstance(X, (int, Integer)):
        return Gamma0(X).dimension_new_cusp_forms(k, p=p)
    raise TypeError(f"X (={X}) must be an integer, a Dirichlet character or a congruence subgroup of type Gamma0, Gamma1 or GammaH")


def dimension_cusp_forms(X, k=2):
    r"""
    The dimension of the space of cusp forms for the given congruence
    subgroup or Dirichlet character.

    INPUT:

    - ``X`` -- congruence subgroup or Dirichlet character
      or integer

    - ``k`` -- weight (integer)

    EXAMPLES::

        sage: from sage.modular.dims import dimension_cusp_forms
        sage: dimension_cusp_forms(5,4)
        1

        sage: dimension_cusp_forms(Gamma0(11),2)
        1
        sage: dimension_cusp_forms(Gamma1(13),2)
        2

        sage: dimension_cusp_forms(DirichletGroup(13).0^2,2)
        1
        sage: dimension_cusp_forms(DirichletGroup(13).0,3)
        1

        sage: dimension_cusp_forms(Gamma0(11),2)
        1
        sage: dimension_cusp_forms(Gamma0(11),0)
        0
        sage: dimension_cusp_forms(Gamma0(1),12)
        1
        sage: dimension_cusp_forms(Gamma0(1),2)
        0
        sage: dimension_cusp_forms(Gamma0(1),4)
        0

        sage: dimension_cusp_forms(Gamma0(389),2)
        32
        sage: dimension_cusp_forms(Gamma0(389),4)
        97
        sage: dimension_cusp_forms(Gamma0(2005),2)
        199
        sage: dimension_cusp_forms(Gamma0(11),1)
        0

        sage: dimension_cusp_forms(Gamma1(11),2)
        1
        sage: dimension_cusp_forms(Gamma1(1),12)
        1
        sage: dimension_cusp_forms(Gamma1(1),2)
        0
        sage: dimension_cusp_forms(Gamma1(1),4)
        0

        sage: dimension_cusp_forms(Gamma1(389),2)
        6112
        sage: dimension_cusp_forms(Gamma1(389),4)
        18721
        sage: dimension_cusp_forms(Gamma1(2005),2)
        159201

        sage: dimension_cusp_forms(Gamma1(11),1)
        0

        sage: e = DirichletGroup(13).0
        sage: e.order()
        12
        sage: dimension_cusp_forms(e,2)
        0
        sage: dimension_cusp_forms(e^2,2)
        1

    Check that :issue:`12640` is fixed::

        sage: dimension_cusp_forms(DirichletGroup(1)(1), 12)
        1
        sage: dimension_cusp_forms(DirichletGroup(2)(1), 24)
        5
    """
    if isinstance(X, dirichlet.DirichletCharacter):
        N = X.modulus()
        if N <= 2:
            return Gamma0(N).dimension_cusp_forms(k)
        return Gamma1(N).dimension_cusp_forms(k, X)
    if isinstance(X, ArithmeticSubgroup):
        return X.dimension_cusp_forms(k)
    if isinstance(X, (int, Integer)):
        return Gamma0(X).dimension_cusp_forms(k)
    raise TypeError("argument 1 must be a Dirichlet character, an integer "
                    "or a finite index subgroup of SL2Z")


def dimension_eis(X, k=2):
    """
    The dimension of the space of Eisenstein series for the given
    congruence subgroup.

    INPUT:

    - ``X`` -- congruence subgroup or Dirichlet character
      or integer

    - ``k`` -- integer; weight

    EXAMPLES::

        sage: from sage.modular.dims import dimension_eis
        sage: dimension_eis(5,4)
        2

        sage: dimension_eis(Gamma0(11),2)
        1
        sage: dimension_eis(Gamma1(13),2)
        11
        sage: dimension_eis(Gamma1(2006),2)
        3711

        sage: e = DirichletGroup(13).0
        sage: e.order()
        12
        sage: dimension_eis(e,2)
        0
        sage: dimension_eis(e^2,2)
        2

        sage: e = DirichletGroup(13).0
        sage: e.order()
        12
        sage: dimension_eis(e,2)
        0
        sage: dimension_eis(e^2,2)
        2
        sage: dimension_eis(e,13)
        2

        sage: G = DirichletGroup(20)
        sage: dimension_eis(G.0,3)
        4
        sage: dimension_eis(G.1,3)
        6
        sage: dimension_eis(G.1^2,2)
        6

        sage: G = DirichletGroup(200)
        sage: e = prod(G.gens(), G(1))
        sage: e.conductor()
        200
        sage: dimension_eis(e,2)
        4

        sage: from sage.modular.dims import dimension_modular_forms
        sage: dimension_modular_forms(Gamma1(4), 11)
        6
    """
    if isinstance(X, ArithmeticSubgroup):
        return X.dimension_eis(k)
    if isinstance(X, dirichlet.DirichletCharacter):
        return Gamma1(X.modulus()).dimension_eis(k, X)
    if isinstance(X, (int, Integer)):
        return Gamma0(X).dimension_eis(k)
    raise TypeError(f"argument in dimension_eis must be an integer, a Dirichlet character, or a finite index subgroup of SL2Z (got {X})")


def dimension_modular_forms(X, k=2):
    r"""
    The dimension of the space of cusp forms for the given congruence
    subgroup (either `\Gamma_0(N)`, `\Gamma_1(N)`, or
    `\Gamma_H(N)`) or Dirichlet character.

    INPUT:

    - ``X`` -- congruence subgroup or Dirichlet character

    - ``k`` -- integer; weight

    EXAMPLES::

        sage: from sage.modular.dims import dimension_modular_forms
        sage: dimension_modular_forms(Gamma0(11),2)
        2
        sage: dimension_modular_forms(Gamma0(11),0)
        1
        sage: dimension_modular_forms(Gamma1(13),2)
        13
        sage: dimension_modular_forms(GammaH(11, [10]), 2)
        10
        sage: dimension_modular_forms(GammaH(11, [10]))
        10
        sage: dimension_modular_forms(GammaH(11, [10]), 4)
        20
        sage: e = DirichletGroup(20).1
        sage: dimension_modular_forms(e,3)
        9
        sage: from sage.modular.dims import dimension_cusp_forms
        sage: dimension_cusp_forms(e,3)
        3
        sage: from sage.modular.dims import dimension_eis
        sage: dimension_eis(e,3)
        6
        sage: dimension_modular_forms(11,2)
        2
    """
    if isinstance(X, (int, Integer)):
        return Gamma0(X).dimension_modular_forms(k)
    if isinstance(X, ArithmeticSubgroup):
        return X.dimension_modular_forms(k)
    if isinstance(X, dirichlet.DirichletCharacter):
        return Gamma1(X.modulus()).dimension_modular_forms(k, eps=X)
    raise TypeError("argument 1 must be an integer, a Dirichlet character "
                    "or an arithmetic subgroup")


def sturm_bound(level, weight=2):
    r"""
    Return the Sturm bound for modular forms with given level and weight.

    For more details, see the documentation for the ``sturm_bound`` method
    of :class:`sage.modular.arithgroup.CongruenceSubgroup` objects.

    INPUT:

    - ``level`` -- integer (interpreted as a level for `\Gamma0`) or a
      congruence subgroup

    - ``weight`` -- integer `\geq 2` (default: 2)

    EXAMPLES::

        sage: from sage.modular.dims import sturm_bound
        sage: sturm_bound(11,2)
        2
        sage: sturm_bound(389,2)
        65
        sage: sturm_bound(1,12)
        1
        sage: sturm_bound(100,2)
        30
        sage: sturm_bound(1,36)
        3
        sage: sturm_bound(11)
        2
    """
    if isinstance(level, ArithmeticSubgroup):
        if level.is_congruence():
            return level.sturm_bound(weight)
        raise ValueError("no Sturm bound defined for noncongruence subgroups")
    if isinstance(level, (int, Integer)):
        return Gamma0(level).sturm_bound(weight)



"""
Utilities for interfacing with the standard library's atexit module.
"""

# ****************************************************************************
#       Copyright (C) 2017 Erik M. Bray <erik.bray@lri.fr>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

import atexit


__all__ = ['restore_atexit']


cdef class restore_atexit:
    r"""
    Context manager that restores the state of the atexit module to its
    previous state when exiting the context.

    INPUT:

    - ``run`` -- boolean (default: ``False``); if ``True``, when exiting the
      context (but before restoring the old exit functions), run all
      atexit functions which were added inside the context

    - ``clear`` -- boolean (default: equal to ``run``); if ``True``, clear
      already registered atexit handlers upon entering the context

    .. WARNING::

        The combination ``run=True`` and ``clear=False`` will cause
        already-registered exit functions to be run twice: once when
        exiting the context and again when exiting Python.

    EXAMPLES:

    For this example we will wrap the entire example with
    ``restore_atexit(clear=True)`` so as to start with a fresh atexit
    module state for the sake of the example.

    Note that the function ``atexit._run_exitfuncs()`` runs all registered
    handlers, and then clears the list of handlers, so we can use it to test
    manipulation of the ``atexit`` state::

        sage: import atexit
        sage: from sage.cpython.atexit import restore_atexit
        sage: def handler(*args, **kwargs):
        ....:     import sys
        ....:     # see https://github.com/sagemath/sage/issues/25270#comment:56
        ....:     sys.stdout.write(str((args, kwargs)))
        ....:     sys.stdout.write('\n')
        sage: atexit.register(handler, 1, 2, c=3)
        <function handler at 0x...>
        sage: atexit.register(handler, 4, 5, d=6)
        <function handler at 0x...>
        sage: with restore_atexit(clear=True):
        ....:     atexit._run_exitfuncs()  # Should be none registered
        ....:     atexit.register(handler, 1, 2, c=3)
        ....:     with restore_atexit():
        ....:         atexit._run_exitfuncs()  # Run just registered handler
        ....:     atexit._run_exitfuncs()  # Handler should be run again
        <function handler at 0x...>
        ((1, 2), {'c': 3})
        ((1, 2), {'c': 3})

    We test the ``run`` option::

        sage: with restore_atexit(run=True):
        ....:     # this handler is run when exiting the context
        ....:     _ = atexit.register(handler, 7, 8, e=9)
        ((7, 8), {'e': 9})
        sage: with restore_atexit(clear=False, run=True):
        ....:     # original handlers are run when exiting the context
        ....:     pass
        ((4, 5), {'d': 6})
        ((1, 2), {'c': 3})

    The original handlers are still in place::

        sage: atexit._run_exitfuncs()
        ((4, 5), {'d': 6})
        ((1, 2), {'c': 3})

    TESTS::

        sage: from sage.cpython.atexit import (_get_exithandlers,
        ....:                                  _clear_exithandlers)
        sage: atexit.register(handler, 1, 2, c=3)
        <function handler at 0x...>
        sage: atexit.register(handler, 4, 5, d=6)
        <function handler at 0x...>
        sage: print("Initial exit handlers:\n{}".format(_get_exithandlers()))
        Initial exit handlers:
        [(<function handler at 0x...>, (1, 2), {'c': 3}),
         (<function handler at 0x...>, (4, 5), {'d': 6})]

        sage: with restore_atexit():
        ....:     pass
        sage: print("After restore_atexit:\n{}".format(_get_exithandlers()))
        After restore_atexit:
        [(<function handler at 0x...>, (1, 2), {'c': 3}),
         (<function handler at 0x...>, (4, 5), {'d': 6})]

        sage: with restore_atexit(clear=True):
        ....:     print("Exit handlers in context manager: {}".format(
        ....:           _get_exithandlers()))
        Exit handlers in context manager: []

        sage: print("After restore_atexit with clear=True:\n{}".format(
        ....:       _get_exithandlers()))
        After restore_atexit with clear=True:
        [(<function handler at 0x...>, (1, 2), {'c': 3}),
         (<function handler at 0x...>, (4, 5), {'d': 6})]
        sage: _clear_exithandlers()
        sage: _get_exithandlers()
        []
    """

    cdef list _exithandlers
    cdef bint _clear, _run

    def __init__(self, *, run=False, clear=None):
        self._clear = self._run = run
        if clear is not None:
            self._clear = clear
        self._exithandlers = None

    def __enter__(self):
        self._exithandlers = _get_exithandlers()
        if self._clear:
            _clear_exithandlers()

        return self

    def __exit__(self, *exc):
        if self._run:
            atexit._run_exitfuncs()
        _set_exithandlers(self._exithandlers)

from cpython.ref cimport PyObject
import sys

# Implement a uniform interface for getting atexit callbacks
cdef extern from *:
    """
    #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE
    #endif
    #undef _PyGC_FINALIZED
    #include "internal/pycore_interp.h"
    #include "internal/pycore_pystate.h"
    
    // Always define this struct for Cython's use
    typedef struct {
        PyObject *func;
        PyObject *args;
        PyObject *kwargs;
    } atexit_callback_struct;
    
    #if PY_VERSION_HEX >= 0x030e0000
    // Python 3.14+: atexit uses a PyList stored in state->callbacks
    // Note: In Python 3.14 the atexit_state struct changed - callbacks is now a PyObject* (PyList)
    
    static PyObject* get_atexit_callbacks_list(PyObject *self) {
        PyInterpreterState *interp = _PyInterpreterState_GET();
        // Access the callbacks list directly from the interpreter state
        // We return a new reference because Cython expects an owned reference
        PyObject *callbacks = interp->atexit.callbacks;
        Py_XINCREF(callbacks);
        return callbacks;
    }
    
    // Dummy function for Python 3.14+ (never called)
    static atexit_callback_struct** get_atexit_callbacks_array(PyObject *self) {
        PyErr_SetString(PyExc_RuntimeError, "Python >= 3.14 has no atexit arrays");
        return NULL;
    }
    #else
    // Python < 3.14: atexit uses C array
    static atexit_callback_struct** get_atexit_callbacks_array(PyObject *self) {
        PyInterpreterState *interp = _PyInterpreterState_GET();
        struct atexit_state state = interp->atexit;
        // Cast from atexit_callback** to our struct type
        return (atexit_callback_struct**)state.callbacks;
    }
    
    // Dummy function for Python < 3.14 (never called)
    static PyObject* get_atexit_callbacks_list(PyObject *self) {
        PyErr_SetString(PyExc_RuntimeError, "Python < 3.14 has no atexit lists");
        return NULL;
    }
    #endif
    """
    # Declare both functions - they exist in all Python versions (one is dummy)
    object get_atexit_callbacks_list(object module)
    
    ctypedef struct atexit_callback_struct:
        PyObject* func
        PyObject* args
        PyObject* kwargs
    atexit_callback_struct** get_atexit_callbacks_array(object module) except NULL


def _get_exithandlers():
    """Return list of exit handlers registered with the atexit module."""
    cdef list exithandlers = []
    cdef atexit_callback_struct ** callbacks
    cdef atexit_callback_struct callback
    cdef int idx
    cdef object kwargs
    
    # Python 3.14+ uses a PyList directly
    if sys.version_info >= (3, 14):
        callbacks_list = get_atexit_callbacks_list(atexit)
        if callbacks_list is None:
            return exithandlers
        # callbacks is a list of tuples: [(func, args, kwargs), ...]
        # Normalize kwargs to ensure it's always a dict (not None)
        # Note: In Python 3.14+, atexit stores callbacks in LIFO order
        # (most recently registered first), but we return them in FIFO
        # order (registration order) for consistency with earlier versions
        for item in reversed(callbacks_list):
            func, args, kwargs = item
            if kwargs is None:
                kwargs = {}
            exithandlers.append((func, args, kwargs))
    else:
        # Python < 3.14 uses C array
        callbacks = get_atexit_callbacks_array(atexit)
        for idx in range(atexit._ncallbacks()):
            callback = callbacks[idx][0]
            if callback.kwargs:
                kwargs = <object>callback.kwargs
            else:
                kwargs = {}
            exithandlers.append((<object>callback.func,
                                 <object>callback.args,
                                 kwargs))
    return exithandlers


def _set_exithandlers(exithandlers):
    """
    Replace the list of exit handlers registered with the atexit module
    with a new list.
    """

    # Clear the existing list
    atexit._clear()

    # We could do this more efficiently by directly rebuilding the array
    # of atexit_callbacks, but this is much simpler
    # Note: exithandlers is in registration order (FIFO).
    # In Python 3.14+, atexit.register prepends to the list (LIFO),
    # so registering in forward order gives us the correct execution order.
    for callback in exithandlers:
        atexit.register(callback[0], *callback[1], **callback[2])


def _clear_exithandlers():
    """Clear the atexit module of all registered exit handlers."""

    atexit._clear()



"""
Sparse rational matrices

AUTHORS:

- William Stein (2007-02-21)
- Soroosh Yazdani (2007-02-21)

TESTS::

    sage: a = matrix(QQ,2,range(4), sparse=True)
    sage: TestSuite(a).run()
    sage: matrix(QQ,0,0,sparse=True).inverse()
    []
"""

#*****************************************************************************
#       Copyright (C) 2007 William Stein <wstein@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from cysignals.signals cimport sig_on, sig_off
from cysignals.memory cimport check_calloc, sig_free

from sage.data_structures.binary_search cimport *
from sage.modules.vector_integer_sparse cimport *
from sage.modules.vector_rational_sparse cimport *

from cpython.sequence cimport *

from sage.rings.rational cimport Rational
from sage.rings.integer  cimport Integer
from sage.matrix.matrix cimport Matrix
from sage.matrix.args cimport SparseEntry, MatrixArgs_init

from sage.libs.gmp.mpz cimport *
from sage.libs.gmp.mpq cimport *

cimport sage.libs.linbox.givaro as givaro
cimport sage.libs.linbox.linbox as linbox
from sage.libs.linbox.conversion cimport new_linbox_matrix_rational_sparse

from sage.libs.flint.fmpq cimport fmpq_set_mpq
from sage.libs.flint.fmpq_mat cimport fmpq_mat_entry

from sage.rings.integer_ring import ZZ
from sage.rings.rational_field import QQ

cimport sage.structure.element

import sage.matrix.matrix_space

from sage.matrix.matrix_integer_sparse cimport Matrix_integer_sparse
from sage.matrix.matrix_rational_dense cimport Matrix_rational_dense


cdef class Matrix_rational_sparse(Matrix_sparse):
    def __cinit__(self):
        self._matrix = <mpq_vector*>check_calloc(self._nrows, sizeof(mpq_vector))
        # initialize the rows
        cdef Py_ssize_t i
        for i in range(self._nrows):
            mpq_vector_init(&self._matrix[i], self._ncols, 0)

    def __dealloc__(self):
        cdef Py_ssize_t i
        if self._matrix is not NULL:
            for i in range(self._nrows):
                mpq_vector_clear(&self._matrix[i])
            sig_free(self._matrix)

    def __init__(self, parent, entries=None, copy=None, bint coerce=True):
        r"""
        Create a sparse matrix over the rational numbers.

        INPUT:

        - ``parent`` -- a matrix space over `\QQ`

        - ``entries`` -- see :func:`matrix`

        - ``copy`` -- ignored (for backwards compatibility)

        - ``coerce`` -- if ``False``, assume without checking that the
          entries are of type :class:`Rational`
        """
        ma = MatrixArgs_init(parent, entries)
        cdef Rational z
        for t in ma.iter(coerce, True):
            se = <SparseEntry>t
            z = <Rational>se.entry
            if z:
                mpq_vector_set_entry(&self._matrix[se.i], se.j, z.value)

    cdef set_unsafe(self, Py_ssize_t i, Py_ssize_t j, x):
        mpq_vector_set_entry(&self._matrix[i], j, (<Rational> x).value)

    cdef get_unsafe(self, Py_ssize_t i, Py_ssize_t j):
        cdef Rational x
        x = Rational()
        mpq_vector_get_entry(x.value, &self._matrix[i], j)
        return x

    cdef copy_from_unsafe(self, Py_ssize_t iDst, Py_ssize_t jDst, src, Py_ssize_t iSrc, Py_ssize_t jSrc):
        r"""
        Copy the ``(iSrc, jSrc)`` entry of ``src`` into the ``(iDst, jDst)``
        entry of ``self``.

        INPUT:

        - ``iDst`` - the row to be copied to in ``self``.
        - ``jDst`` - the column to be copied to in ``self``.
        - ``src`` - the matrix to copy from. Should be a Matrix_rational_sparse
                    with the same base ring as ``self``.
        - ``iSrc``  - the row to be copied from in ``src``.
        - ``jSrc`` - the column to be copied from in ``src``.

        TESTS::

            sage: M = matrix(QQ,3,4,[i + 1/(i+1) if is_prime(i) else 0 for i in range(12)],sparse=True)
            sage: M
            [     0      0    7/3   13/4]
            [     0   31/6      0   57/8]
            [     0      0      0 133/12]
            sage: M.transpose()
            [     0      0      0]
            [     0   31/6      0]
            [   7/3      0      0]
            [  13/4   57/8 133/12]
            sage: M.matrix_from_rows([0,2])
            [     0      0    7/3   13/4]
            [     0      0      0 133/12]
            sage: M.matrix_from_columns([1,3])
            [     0   13/4]
            [  31/6   57/8]
            [     0 133/12]
            sage: M.matrix_from_rows_and_columns([1,2],[0,3])
            [     0   57/8]
            [     0 133/12]
        """
        cdef Rational x
        x = Rational()
        cdef Matrix_rational_sparse _src = <Matrix_rational_sparse> src
        mpq_vector_get_entry(x.value, &_src._matrix[iSrc], jSrc)
        mpq_vector_set_entry(&self._matrix[iDst], jDst, x.value)

    cdef bint get_is_zero_unsafe(self, Py_ssize_t i, Py_ssize_t j) except -1:
        """
        Return 1 if the entry ``(i, j)`` is zero, otherwise 0.

        EXAMPLES::

            sage: M = matrix(QQ, [[0,1,0],[0,0,0]], sparse=True)
            sage: M.zero_pattern_matrix()  # indirect doctest
            [1 0 1]
            [1 1 1]
        """
        return mpq_vector_is_entry_zero_unsafe(&self._matrix[i], j)

    def add_to_entry(self, Py_ssize_t i, Py_ssize_t j, elt):
        r"""
        Add ``elt`` to the entry at position ``(i, j)``.

        EXAMPLES::

            sage: m = matrix(QQ, 2, 2, sparse=True)
            sage: m.add_to_entry(0, 0, -1/3)
            sage: m
            [-1/3    0]
            [   0    0]
            sage: m.add_to_entry(0, 0, 1/3)
            sage: m
            [0 0]
            [0 0]
            sage: m.nonzero_positions()
            []
        """
        if not isinstance(elt, Rational):
            elt = Rational(elt)
        if i < 0:
            i += self._nrows
        if i < 0 or i >= self._nrows:
            raise IndexError("row index out of range")
        if j < 0:
            j += self._ncols
        if j < 0 or j >= self._ncols:
            raise IndexError("column index out of range")

        cdef mpq_t z
        mpq_init(z)
        mpq_vector_get_entry(z, &self._matrix[i], j)
        mpq_add(z, z, (<Rational>elt).value)
        mpq_vector_set_entry(&self._matrix[i], j, z)
        mpq_clear(z)

    ########################################################################
    # LEVEL 2 functionality
    #   * def _pickle
    #   * def _unpickle
    #   * cdef _add_
    #   * cdef _sub_
    #   * cdef _mul_
    #   * cpdef _richcmp_
    #   * __neg__
    #   * __invert__
    #   * __copy__
    #   * _multiply_classical
    #   * _matrix_times_matrix_
    #   * _list -- list of underlying elements (need not be a copy)
    #   * x _dict -- sparse dictionary of underlying elements (need not be a copy)

    cdef sage.structure.element.Matrix _matrix_times_matrix_(self, sage.structure.element.Matrix _right):
        cdef Matrix_rational_sparse right, ans
        right = _right

        cdef mpq_vector* v

        # Build a table that gives the nonzero positions in each column of right
        cdef list nonzero_positions_in_columns = [set() for _ in range(right._ncols)]
        cdef Py_ssize_t i, j, k
        for i in range(right._nrows):
            v = &(right._matrix[i])
            for j in range(v.num_nonzero):
                (<set> nonzero_positions_in_columns[v.positions[j]]).add(i)
        # pre-computes the list of nonzero columns of right
        cdef list right_indices
        right_indices = [j for j in range(right._ncols)
                         if nonzero_positions_in_columns[j]]

        ans = self.new_matrix(self._nrows, right._ncols)

        # Now do the multiplication, getting each row completely before filling it in.
        cdef set c
        cdef mpq_t x, y, s
        mpq_init(x)
        mpq_init(y)
        mpq_init(s)
        for i in range(self._nrows):
            v = &(self._matrix[i])
            if not v.num_nonzero:
                continue
            for j in right_indices:
                mpq_set_si(s, 0, 1)
                c = <set> nonzero_positions_in_columns[j]
                for k in range(v.num_nonzero):
                    if v.positions[k] in c:
                        mpq_vector_get_entry(y, &right._matrix[v.positions[k]], j)
                        mpq_mul(x, v.entries[k], y)
                        mpq_add(s, s, x)
                mpq_vector_set_entry(&ans._matrix[i], j, s)

        mpq_clear(x)
        mpq_clear(y)
        mpq_clear(s)
        return ans

    def _matrix_times_matrix_dense(self, sage.structure.element.Matrix _right):
        """
        Do the sparse matrix multiply, but return a dense matrix as the result.

        EXAMPLES::

            sage: a = matrix(QQ, 2, [1,2,3,4], sparse=True)
            sage: b = matrix(QQ, 2, 3, [1..6], sparse=True)
            sage: a * b
            [ 9 12 15]
            [19 26 33]
            sage: c = a._matrix_times_matrix_dense(b); c
            [ 9 12 15]
            [19 26 33]
            sage: type(c)
            <class 'sage.matrix.matrix_rational_dense.Matrix_rational_dense'>
        """
        cdef Matrix_rational_sparse right
        cdef Matrix_rational_dense ans
        right = _right

        cdef mpq_vector* v

        # Build a table that gives the nonzero positions in each column of right
        nonzero_positions_in_columns = [set() for _ in range(right._ncols)]
        cdef Py_ssize_t i, j, k
        for i in range(right._nrows):
            v = &(right._matrix[i])
            for j in range(right._matrix[i].num_nonzero):
                nonzero_positions_in_columns[v.positions[j]].add(i)

        ans = self.new_matrix(self._nrows, right._ncols, sparse=False)

        # Now do the multiplication, getting each row completely before filling it in.
        cdef mpq_t x, y, s
        mpq_init(x)
        mpq_init(y)
        mpq_init(s)
        for i in range(self._nrows):
            v = &self._matrix[i]
            for j in range(right._ncols):
                mpq_set_si(s, 0, 1)
                c = nonzero_positions_in_columns[j]
                for k in range(v.num_nonzero):
                    if v.positions[k] in c:
                        mpq_vector_get_entry(y, &right._matrix[v.positions[k]], j)
                        mpq_mul(x, v.entries[k], y)
                        mpq_add(s, s, x)
                fmpq_set_mpq(fmpq_mat_entry(ans._matrix, i, j), s)

        mpq_clear(x)
        mpq_clear(y)
        mpq_clear(s)
        return ans

    ########################################################################
    # def _pickle(self):
    # def _unpickle(self, data, int version):   # use version >= 0
    # cpdef _add_(self, right):
    # cdef _mul_(self, Matrix right):
    # cpdef _richcmp_(self, Matrix right, int op):
    # def __neg__(self):
    # def __invert__(self):
    # def __copy__(self):
    # def _multiply_classical(left, matrix.Matrix _right):
    # def _list(self):

    # TODO
    ##     cpdef _lmul_(self, Element right):
    ##         """
    ##         EXAMPLES::
    ##
    ##             sage: a = matrix(QQ,2,range(6))
    ##             sage: (3/4) * a
    ##             [   0  3/4  3/2]
    ##             [ 9/4    3 15/4]
    ##         """

    def _dict(self):
        """
        Unsafe version of the dict method, mainly for internal use.

        This may return the dict of elements, but as an *unsafe*
        reference to the underlying dict of the object.  It might
        be dangerous if you change entries of the returned dict.
        """
        d = self.fetch('dict')
        if d is not None:
            return d

        cdef Py_ssize_t i, j
        d = {}
        for i in range(self._nrows):
            for j in range(self._matrix[i].num_nonzero):
                x = Rational()
                mpq_set((<Rational>x).value, self._matrix[i].entries[j])
                d[(int(i), int(self._matrix[i].positions[j]))] = x
        self.cache('dict', d)
        return d

    ########################################################################
    # LEVEL 3 functionality (Optional)
    #    * cdef _sub_
    #    * __deepcopy__
    #    * __invert__
    #    * Matrix windows -- only if you need strassen for that base
    #    * Other functions (list them here):
    ########################################################################

    def _nonzero_positions_by_row(self, copy=True):
        """
        Return the list of pairs (i,j) such that self[i,j] != 0.

        It is safe to change the resulting list (unless you give the option copy=False).

        EXAMPLES::

            sage: M = Matrix(QQ, [[0,0,0,1,0,0,0,0],[0,1,0,0,0,0,1,0]], sparse=True); M
            [0 0 0 1 0 0 0 0]
            [0 1 0 0 0 0 1 0]
            sage: M.nonzero_positions()
            [(0, 3), (1, 1), (1, 6)]
        """
        x = self.fetch('nonzero_positions')
        if x is not None:
            if copy:
                return list(x)
            return x
        nzp = []
        cdef Py_ssize_t i, j
        for i from 0 <= i < self._nrows:
            for j from 0 <= j < self._matrix[i].num_nonzero:
                nzp.append((i,self._matrix[i].positions[j]))
        self.cache('nonzero_positions', nzp)
        if copy:
            return list(nzp)
        return nzp

    def height(self):
        """
        Return the height of this matrix, which is the least common
        multiple of all numerators and denominators of elements of
        this matrix.

        OUTPUT: integer

        EXAMPLES::

            sage: b = matrix(QQ,2,range(6), sparse=True); b[0,0]=-5007/293; b
            [-5007/293         1         2]
            [        3         4         5]
            sage: b.height()
            5007
        """
        cdef Integer z = Integer.__new__(Integer)
        self.mpz_height(z.value)
        return z

    cdef int mpz_height(self, mpz_t height) except -1:
        cdef mpz_t x, h
        mpz_init(x)
        mpz_init_set_si(h, 0)
        cdef Py_ssize_t i, j
        sig_on()
        for i from 0 <= i < self._nrows:
            for j from 0 <= j < self._matrix[i].num_nonzero:
                mpq_get_num(x, self._matrix[i].entries[j])
                mpz_abs(x, x)
                if mpz_cmp(h,x) < 0:
                    mpz_set(h,x)
                mpq_get_den(x, self._matrix[i].entries[j])
                mpz_abs(x, x)
                if mpz_cmp(h,x) < 0:
                    mpz_set(h,x)
        sig_off()
        mpz_set(height, h)
        mpz_clear(h)
        mpz_clear(x)
        return 0

    cdef int mpz_denom(self, mpz_t d) except -1:
        mpz_set_si(d, 1)
        cdef Py_ssize_t i, j

        sig_on()
        for i in range(self._nrows):
            for j in range(self._matrix[i].num_nonzero):
                mpz_lcm(d, d, mpq_denref(self._matrix[i].entries[j]))
        sig_off()
        return 0

    def denominator(self):
        """
        Return the denominator of this matrix.

        OUTPUT: Sage Integer

        EXAMPLES::

            sage: b = matrix(QQ,2,range(6)); b[0,0]=-5007/293; b
            [-5007/293         1         2]
            [        3         4         5]
            sage: b.denominator()
            293
        """
        cdef Integer z = Integer.__new__(Integer)
        self.mpz_denom(z.value)
        return z

    def _clear_denom(self):
        """
        OUTPUT: ``d*self, D``

        The product D*self is a matrix over ZZ

        EXAMPLES::

            sage: a = matrix(QQ,3,[-2/7, -1/4, -2, 0, 1/7, 1, 0, 1/2, 1/5],sparse=True)
            sage: a.denominator()
            140
            sage: a._clear_denom()
            (
            [ -40  -35 -280]
            [   0   20  140]
            [   0   70   28], 140
            )
        """
        cdef Integer D
        cdef Py_ssize_t i, j
        cdef Matrix_integer_sparse A
        cdef mpz_t t
        cdef mpq_vector* v

        D = Integer()
        self.mpz_denom(D.value)

        MZ = sage.matrix.matrix_space.MatrixSpace(ZZ, self._nrows, self._ncols, sparse=True)
        A = MZ.zero_matrix().__copy__()

        mpz_init(t)
        sig_on()
        for i from 0 <= i < self._nrows:
            v = &(self._matrix[i])
            for j from 0 <= j < v.num_nonzero:
                mpz_divexact(t, D.value, mpq_denref(v.entries[j]))
                mpz_mul(t, t, mpq_numref(v.entries[j]))
                mpz_vector_set_entry(&(A._matrix[i]), v.positions[j], t)
        sig_off()
        mpz_clear(t)
        return A, D

    ################################################
    # Echelon form
    ################################################
    def echelonize(self, height_guess=None, proof=True, **kwds):
        """
        Transform the matrix ``self`` into reduced row echelon form
        in place.

        INPUT:

        - ``height_guess``, ``proof``, ``**kwds`` -- all passed to the multimodular
          algorithm; ignored by the `p`-adic algorithm

        OUTPUT:

        Nothing. The matrix ``self`` is transformed into reduced row
        echelon form in place.

        ALGORITHM: a multimodular algorithm.

        EXAMPLES::

            sage: a = matrix(QQ, 4, range(16), sparse=True); a[0,0] = 1/19; a[0,1] = 1/5; a
            [1/19  1/5    2    3]
            [   4    5    6    7]
            [   8    9   10   11]
            [  12   13   14   15]
            sage: a.echelonize(); a
            [      1       0       0 -76/157]
            [      0       1       0  -5/157]
            [      0       0       1 238/157]
            [      0       0       0       0]

        :issue:`10319` has been fixed::

            sage: m = Matrix(QQ, [1], sparse=True); m.echelonize()
            sage: m = Matrix(QQ, [1], sparse=True); m.echelonize(); m
            [1]
        """

        x = self.fetch('in_echelon_form')
        if x is not None:
            return  # already known to be in echelon form
        self.check_mutability()

        pivots = self._echelonize_multimodular(height_guess, proof, **kwds)

        self.cache('in_echelon_form', True)
        self.cache('echelon_form', self)
        self.cache('pivots', pivots)
        self.cache('rank', len(pivots))

    def echelon_form(self, algorithm='default',
                     height_guess=None, proof=True, **kwds):
        """
        INPUT:

        - ``height_guess``, ``proof``, ``**kwds`` -- all passed to the multimodular
          algorithm; ignored by the `p`-adic algorithm

        OUTPUT: ``self`` is no in reduced row echelon form

        EXAMPLES::

            sage: a = matrix(QQ, 4, range(16), sparse=True); a[0,0] = 1/19; a[0,1] = 1/5; a
            [1/19  1/5    2    3]
            [   4    5    6    7]
            [   8    9   10   11]
            [  12   13   14   15]
            sage: a.echelon_form()
            [      1       0       0 -76/157]
            [      0       1       0  -5/157]
            [      0       0       1 238/157]
            [      0       0       0       0]
        """
        label = 'echelon_form_%s' % algorithm
        x = self.fetch(label)
        if x is not None:
            return x
        if self.fetch('in_echelon_form'):
            return self

        E, pivots = self._echelon_form_multimodular(height_guess, proof=proof)

        self.cache(label, E)
        self.cache('pivots', pivots)
        return E

    # Multimodular echelonization algorithms
    def _echelonize_multimodular(self, height_guess=None, proof=True, **kwds):
        cdef Matrix_rational_sparse E
        E, pivots = self._echelon_form_multimodular(height_guess, proof=proof, **kwds)
        self.clear_cache()
        # Swap the data of E and self (effectively moving E to self)
        self._matrix, E._matrix = E._matrix, self._matrix
        return pivots

    def _echelon_form_multimodular(self, height_guess=None, proof=True):
        """
        Return reduced row-echelon form using a multi-modular
        algorithm.  Does not change ``self``.

        INPUT:

        - ``height_guess`` -- integer or ``None``
        - ``proof`` -- boolean (default: ``True``)
        """
        from sage.matrix.misc import matrix_rational_echelon_form_multimodular
        cdef Matrix E
        E, pivots = matrix_rational_echelon_form_multimodular(self,
                                 height_guess=height_guess, proof=proof)
        E._parent = self._parent
        return E, pivots

    def set_row_to_multiple_of_row(self, i, j, s):
        """
        Set row i equal to s times row j.

        EXAMPLES::

            sage: a = matrix(QQ,2,3,range(6), sparse=True); a
            [0 1 2]
            [3 4 5]
            sage: a.set_row_to_multiple_of_row(1,0,-3)
            sage: a
            [ 0  1  2]
            [ 0 -3 -6]
        """
        self.check_row_bounds_and_mutability(i, j)
        cdef Rational _s
        _s = Rational(s)
        mpq_vector_scalar_multiply(&self._matrix[i], &self._matrix[j], _s.value)

    def dense_matrix(self):
        """
        Return dense version of this matrix.

        EXAMPLES::

            sage: a = matrix(QQ,2,[1..4],sparse=True); type(a)
            <class 'sage.matrix.matrix_rational_sparse.Matrix_rational_sparse'>
            sage: type(a.dense_matrix())
            <class 'sage.matrix.matrix_rational_dense.Matrix_rational_dense'>
            sage: a.dense_matrix()
            [1 2]
            [3 4]

        Check that subdivisions are preserved when converting between
        dense and sparse matrices::

            sage: a.subdivide([1,1], [2])
            sage: b = a.dense_matrix().sparse_matrix().dense_matrix()
            sage: b.subdivisions() == a.subdivisions()
            True
        """
        cdef Matrix_rational_dense B
        cdef mpq_vector* v

        B = self.matrix_space(sparse=False).zero_matrix().__copy__()
        for i from 0 <= i < self._nrows:
            v = &(self._matrix[i])
            for j from 0 <= j < v.num_nonzero:
                fmpq_set_mpq(fmpq_mat_entry(B._matrix, i, v.positions[j]), v.entries[j])
        B.subdivide(self.subdivisions())
        return B

##     def _set_row_to_negative_of_row_of_A_using_subset_of_columns(self, Py_ssize_t i, Matrix A,
##                                                                  Py_ssize_t r, cols):
##         B = self.__copy__()
##         B.x_set_row_to_negative_of_row_of_A_using_subset_of_columns(i, A, r, cols)
##         cdef Py_ssize_t l
##         l = 0
##         for z in range(self.ncols()):
##             self[i,z] = 0
##         for k in cols:
##             self.set_unsafe(i,l,-A.get_unsafe(r,k))               #self[i,l] = -A[r,k]
##             l += 1
##         if self != B:
##             print("correct =\n", self.str())
##             print("wrong = \n", B.str())
##             print("diff = \n", (self-B).str())

    def _set_row_to_negative_of_row_of_A_using_subset_of_columns(self, Py_ssize_t i, Matrix A,
                                                                 Py_ssize_t r, cols,
                                                                 cols_index=None):
        """
        Set row i of ``self`` to -(row r of A), but where we only take the
        given column positions in that row of A.  Note that we *DO*
        zero out the other entries of ``self``'s row i.

        INPUT:

        - ``i`` -- integer, index into the rows of self
        - ``A`` -- a sparse matrix
        - ``r`` -- integer, index into rows of A
        - ``cols`` -- a *sorted* list of integers
        - ``cols_index`` -- (optional) set it to this to vastly speed up
          calls to this function::

                dict([(cols[i], i) for i in range(len(cols))])

        EXAMPLES::

            sage: a = matrix(QQ,2,3,range(6), sparse=True); a
            [0 1 2]
            [3 4 5]

        Note that the row is zeroed out before being set in the sparse case. ::

            sage: a._set_row_to_negative_of_row_of_A_using_subset_of_columns(0,a,1,[1,2])
            sage: a
            [-4 -5  0]
            [ 3  4  5]
        """
        # this function exists just because it is useful for modular symbols presentations.
        self.check_row_bounds_and_mutability(i,i)
        if r < 0 or r >= A.nrows():
            raise IndexError("invalid row")

        if not A.is_sparse():
            A = A.sparse_matrix()

        if A.base_ring() != QQ:
            A = A.change_ring(QQ)

        cdef Matrix_rational_sparse _A
        _A = A

        cdef Py_ssize_t l, n

        cdef mpq_vector *v
        cdef mpq_vector *w
        v = &self._matrix[i]
        w = &_A._matrix[r]

        if cols_index is None:
            cols_index = dict([(cols[i], i) for i in range(len(cols))])

        _cols = set(cols)
        pos = [i for i from 0 <= i < w.num_nonzero if w.positions[i] in _cols]
        n = len(pos)

        mpq_vector_clear(v)
        allocate_mpq_vector(v, n)
        v.num_nonzero = n
        v.degree = self._ncols

        for l from 0 <= l < n:
            v.positions[l] = cols_index[w.positions[pos[l]]]
            mpq_mul(v.entries[l], w.entries[pos[l]], minus_one)

    def _right_kernel_matrix(self, algorithm='default', proof=None):
        r"""
        Return a pair that includes a matrix of basis vectors
        for the right kernel of ``self``.

        INPUT:

        - ``algorithm`` -- (default: ``'default'``) a keyword that selects the
          algorithm employed.  Allowable values are:

          - ``'default'`` -- equivalent to ``'padic'``
          - ``'padic'`` -- `p`-adic algorithm from the IML library for matrices
            over the rationals and integers
          - ``'linbox'`` -- LinBox library code for sparse matrices over the
            rationals

        OUTPUT:

        Returns a pair.  First item is a string that identifies the nature
        of the basis vectors, either 'computed-iml-rational' or
        'computed-linbox-rational'.

        Second item is a matrix whose nonzero rows are a basis for the right kernel,
        over the rationals, as computed by the IML library or the LinBox library.

        EXAMPLES::

            sage: A = matrix(QQ, [
            ....:                 [1, 0, 1, -3, 1],
            ....:                 [-5, 1, 0, 7, -3],
            ....:                 [0, -1, -4, 6, -2],
            ....:                 [4, -1, 0, -6, 2]],
            ....:             sparse=True)
            sage: result = A._right_kernel_matrix()
            sage: result[0]
            'computed-iml-rational'
            sage: result[1]
            [-1  2 -2 -1  0]
            [ 1  2  0  0 -1]
            sage: X = result[1].transpose()
            sage: A*X == zero_matrix(QQ, 4, 2)
            True
            sage: result = A._right_kernel_matrix(algorithm='linbox')
            sage: result[0]
            'computed-linbox-rational'
            sage: result[1]
            [ 1 -2  2  1  0]
            [-1 -2  0  0  1]
            sage: X = result[1].transpose()
            sage: A*X == zero_matrix(QQ, 4, 2)
            True

        TESTS:

        We test three trivial cases. ::

            sage: A = matrix(QQ, 0, 2, sparse=True)
            sage: A._right_kernel_matrix()[1]
            [1 0]
            [0 1]
            sage: A = matrix(QQ, 2, 0, sparse=True)
            sage: A._right_kernel_matrix()[1].parent()
            Full MatrixSpace of 0 by 0 dense matrices over Rational Field
            sage: A = zero_matrix(QQ, 4, 3,  sparse=True)
            sage: A._right_kernel_matrix()[1]
            [1 0 0]
            [0 1 0]
            [0 0 1]

        .. SEEALSO::

            :meth:`~sage.matrix.matrix_rational_dense.Matrix_rational_dense._right_kernel_matrix`
        """
        if algorithm == 'default' or algorithm == 'padic':
            return self.dense_matrix()._right_kernel_matrix()
        elif algorithm == 'linbox':
            return self._right_kernel_matrix_linbox()
        else:
            raise ValueError(f"matrix kernel algorithm '{algorithm}' not recognized")

    def _right_kernel_matrix_linbox(self):
        r"""
        Return a pair that includes a matrix of basis vectors
        for the right kernel of ``self``.

        OUTPUT:

        Returns a pair.  First item is the string 'computed-linbox-rational'
        that identifies the nature of the basis vectors.

        Second item is a matrix whose rows are a basis for the right kernel,
        over the rationals, as computed by the LinBox library.
        """
        # NOTE: Degenerate cases (0 rows or columns) are handled by right_kernel_matrix.

        cdef givaro.QField givQQ
        cdef linbox.SparseMatrix_rational * M = new_linbox_matrix_rational_sparse(givQQ, self)

        MQ = sage.matrix.matrix_space.MatrixSpace(QQ, self._ncols, self._ncols, sparse=True)
        A = MQ.zero_matrix().__copy__()

        cdef linbox.SparseMatrix_rational * N = new_linbox_matrix_rational_sparse(givQQ, A)

        cdef linbox.GaussDomain_rational * dom = new linbox.GaussDomain_rational(givQQ)

        cdef Matrix_rational_sparse ans
        cdef mpq_t s

        try:
            sig_on()
            dom.nullspacebasisin(N[0], M[0])
            sig_off()
        except KeyboardInterrupt:
            del N
            raise
        finally:
            del M, dom

        ans = self.new_matrix(N.coldim(), N.rowdim()) # NOTE: Transposing.
        cdef size_t i, k
        cdef Py_ssize_t j
        mpq_init(s)
        for i in range(N.rowdim()):
            for k in range(N.getRow(i).size()):
                j = <Py_ssize_t> N.getRow(i)[k].first
                entry = N.getRow(i)[k].second
                mpq_set_num(s, entry.nume().get_mpz())
                mpq_set_den(s, entry.deno().get_mpz())
                mpq_vector_set_entry(&ans._matrix[j], i, s) # NOTE: Transposing.
        mpq_clear(s)

        del N

        # Remove zero rows, if any.
        if ans.nrows() > 0:
            r = ans.nrows() - 1
            while r >= 0 and ans.row(r) == 0:
                r -= 1
            ans = ans.submatrix(0, 0, r + 1)

        return 'computed-linbox-rational', ans


#########################

# used for a function above
cdef mpq_t minus_one
mpq_init(minus_one)
mpq_set_si(minus_one, -1,1)



r"""
Trivial valuations

AUTHORS:

- Julian Rüth (2016-10-14): initial version

EXAMPLES::

    sage: v = valuations.TrivialValuation(QQ); v
    Trivial valuation on Rational Field
    sage: v(1)
    0
"""
# ****************************************************************************
#       Copyright (C) 2016-2017 Julian Rüth <julian.rueth@fsfe.org>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from .valuation import DiscretePseudoValuation, DiscreteValuation, InfiniteDiscretePseudoValuation
from .valuation_space import DiscretePseudoValuationSpace
from sage.structure.factory import UniqueFactory


class TrivialValuationFactory(UniqueFactory):
    r"""
    Create a trivial valuation on ``domain``.

    EXAMPLES::

        sage: v = valuations.TrivialValuation(QQ); v
        Trivial valuation on Rational Field
        sage: v(1)
        0
    """
    def __init__(self, clazz, parent, *args, **kwargs):
        r"""
        TESTS::

            sage: from sage.rings.valuation.trivial_valuation import TrivialValuationFactory
            sage: isinstance(valuations.TrivialValuation, TrivialValuationFactory)
            True
        """
        UniqueFactory.__init__(self, *args, **kwargs)
        self._class = clazz
        self._parent = parent

    def create_key(self, domain):
        r"""
        Create a key that identifies this valuation.

        EXAMPLES::

            sage: valuations.TrivialValuation(QQ) is valuations.TrivialValuation(QQ) # indirect doctest
            True
        """
        return domain,

    def create_object(self, version, key, **extra_args):
        r"""
        Create a trivial valuation from ``key``.

        EXAMPLES::

            sage: valuations.TrivialValuation(QQ) # indirect doctest
            Trivial valuation on Rational Field
        """
        domain, = key
        parent = self._parent(domain)
        return parent.__make_element_class__(self._class)(parent)


class TrivialDiscretePseudoValuation_base(DiscretePseudoValuation):
    r"""
    Base class for code shared by trivial valuations.

    EXAMPLES::

        sage: v = valuations.TrivialPseudoValuation(ZZ); v
        Trivial pseudo-valuation on Integer Ring

    TESTS::

        sage: TestSuite(v).run() # long time
    """
    def uniformizer(self):
        r"""
        Return a uniformizing element for this valuation.

        EXAMPLES::

            sage: v = valuations.TrivialPseudoValuation(ZZ)
            sage: v.uniformizer()
            Traceback (most recent call last):
            ...
            ValueError: Trivial valuations do not define a uniformizing element
        """
        raise ValueError("Trivial valuations do not define a uniformizing element")

    def is_trivial(self):
        r"""
        Return whether this valuation is trivial.

        EXAMPLES::

            sage: v = valuations.TrivialPseudoValuation(QQ)
            sage: v.is_trivial()
            True
        """
        return True

    def is_negative_pseudo_valuation(self):
        r"""
        Return whether this valuation attains the value `-\infty`.

        EXAMPLES::

            sage: v = valuations.TrivialPseudoValuation(QQ)
            sage: v.is_negative_pseudo_valuation()
            False
        """
        return False


class TrivialDiscretePseudoValuation(TrivialDiscretePseudoValuation_base, InfiniteDiscretePseudoValuation):
    r"""
    The trivial pseudo-valuation that is `\infty` everywhere.

    EXAMPLES::

        sage: v = valuations.TrivialPseudoValuation(QQ); v
        Trivial pseudo-valuation on Rational Field

    TESTS::

        sage: TestSuite(v).run() # long time
    """
    def __init__(self, parent):
        r"""
        TESTS::

            sage: from sage.rings.valuation.trivial_valuation import TrivialDiscretePseudoValuation
            sage: v = valuations.TrivialPseudoValuation(QQ)
            sage: isinstance(v, TrivialDiscretePseudoValuation)
            True
        """
        TrivialDiscretePseudoValuation_base.__init__(self, parent)
        InfiniteDiscretePseudoValuation.__init__(self, parent)

    def _call_(self, x):
        r"""
        Evaluate this valuation at ``x``.

        EXAMPLES::

            sage: v = valuations.TrivialPseudoValuation(QQ)
            sage: v(0)
            +Infinity
            sage: v(1)
            +Infinity
        """
        from sage.rings.infinity import infinity
        return infinity

    def _repr_(self):
        r"""
        Return a printable representation of this valuation.

        EXAMPLES::

            sage: valuations.TrivialPseudoValuation(QQ) # indirect doctest
            Trivial pseudo-valuation on Rational Field
        """
        return "Trivial pseudo-valuation on %r" % (self.domain(),)

    def value_group(self):
        r"""
        Return the value group of this valuation.

        EXAMPLES:

        A trivial discrete pseudo-valuation has no value group::

            sage: v = valuations.TrivialPseudoValuation(QQ)
            sage: v.value_group()
            Traceback (most recent call last):
            ...
            ValueError: The trivial pseudo-valuation that is infinity everywhere does not have a value group.
        """
        raise ValueError("The trivial pseudo-valuation that is infinity everywhere does not have a value group.")

    def residue_ring(self):
        r"""
        Return the residue ring of this valuation.

        EXAMPLES::

            sage: valuations.TrivialPseudoValuation(QQ).residue_ring()
            Quotient of Rational Field by the ideal (1)
        """
        return self.domain().quo(self.domain().one())

    def reduce(self, x):
        r"""
        Reduce ``x`` modulo the positive elements of this valuation.

        EXAMPLES::

            sage: v = valuations.TrivialPseudoValuation(QQ)
            sage: v.reduce(1)
            0
        """
        self.domain().coerce(x)
        return self.residue_ring().zero()

    def lift(self, X):
        r"""
        Return a lift of ``X`` to the domain of this valuation.

        EXAMPLES::

            sage: v = valuations.TrivialPseudoValuation(QQ)
            sage: v.lift(v.residue_ring().zero())
            0
        """
        self.residue_ring().coerce(X)  # ignore the output
        return self.domain().zero()

    def _ge_(self, other):
        r"""
        Return whether this valuation is bigger or equal than ``other``
        everywhere.

        EXAMPLES::

            sage: v = valuations.TrivialPseudoValuation(QQ)
            sage: w = valuations.TrivialValuation(QQ)
            sage: v >= w
            True
        """
        # the trivial discrete valuation is the biggest valuation
        return True


class TrivialDiscreteValuation(TrivialDiscretePseudoValuation_base, DiscreteValuation):
    r"""
    The trivial valuation that is zero on nonzero elements.

    EXAMPLES::

        sage: v = valuations.TrivialValuation(QQ); v
        Trivial valuation on Rational Field

    TESTS::

        sage: TestSuite(v).run() # long time
    """
    def __init__(self, parent):
        r"""
        TESTS::

            sage: from sage.rings.valuation.trivial_valuation import TrivialDiscreteValuation
            sage: v = valuations.TrivialValuation(QQ)
            sage: isinstance(v, TrivialDiscreteValuation)
            True
        """
        TrivialDiscretePseudoValuation_base.__init__(self, parent)
        DiscreteValuation.__init__(self, parent)

    def _call_(self, x):
        r"""
        Evaluate this valuation at ``x``.

        EXAMPLES::

            sage: v = valuations.TrivialValuation(QQ)
            sage: v(0)
            +Infinity
            sage: v(1)
            0
        """
        from sage.rings.infinity import infinity
        return infinity if x == 0 else self.codomain().zero()

    def _repr_(self):
        r"""
        Return a printable representation of this valuation.

        EXAMPLES::

            sage: valuations.TrivialValuation(QQ) # indirect doctest
            Trivial valuation on Rational Field
        """
        return "Trivial valuation on %r" % (self.domain(),)

    def value_group(self):
        r"""
        Return the value group of this valuation.

        EXAMPLES:

        A trivial discrete valuation has a trivial value group::

            sage: v = valuations.TrivialValuation(QQ)
            sage: v.value_group()
            Trivial Additive Abelian Group
        """
        from .value_group import DiscreteValueGroup
        return DiscreteValueGroup(0)

    def residue_ring(self):
        r"""
        Return the residue ring of this valuation.

        EXAMPLES::

            sage: valuations.TrivialValuation(QQ).residue_ring()
            Rational Field
        """
        return self.domain()

    def reduce(self, x):
        r"""
        Reduce ``x`` modulo the positive elements of this valuation.

        EXAMPLES::

            sage: v = valuations.TrivialValuation(QQ)
            sage: v.reduce(1)
            1
        """
        return self.domain().coerce(x)

    def lift(self, X):
        r"""
        Return a lift of ``X`` to the domain of this valuation.

        EXAMPLES::

            sage: v = valuations.TrivialValuation(QQ)
            sage: v.lift(v.residue_ring().zero())
            0
        """
        return self.residue_ring().coerce(X)

    def extensions(self, ring):
        r"""
        Return the unique extension of this valuation to ``ring``.

        EXAMPLES::

            sage: v = valuations.TrivialValuation(ZZ)
            sage: v.extensions(QQ)
            [Trivial valuation on Rational Field]
        """
        if self.domain().is_subring(ring):
            return [TrivialValuation(ring)]
        return super(DiscretePseudoValuation, self).extensions(ring)

    def _ge_(self, other):
        r"""
        Return whether this valuation is bigger or equal than ``other``
        everywhere.

        EXAMPLES::

            sage: v = valuations.TrivialPseudoValuation(QQ)
            sage: w = valuations.TrivialValuation(QQ)
            sage: w >= v
            False
        """
        # the trivial discrete valuation is the smallest valuation
        return self is other


TrivialValuation = TrivialValuationFactory(TrivialDiscreteValuation, DiscretePseudoValuationSpace, "sage.rings.valuation.trivial_valuation.TrivialValuation")
TrivialPseudoValuation = TrivialValuationFactory(TrivialDiscretePseudoValuation, DiscretePseudoValuationSpace, "sage.rings.valuation.trivial_valuation.TrivialPseudoValuation")



r"""
Mandelbrot and Julia sets

Plots the Mandelbrot and Julia sets for general polynomial maps in the complex
plane.

The Mandelbrot set is the set of complex numbers `c` for which the map
`f_c(z)` does not diverge when iterated from `z = 0`. This set of complex
numbers can be visualized by plotting each value for `c` in the complex plane.
The Mandelbrot set is often an example of a fractal when plotted in the complex
plane. For general one parameter families of polynomials, the mandelbrot set
is the parameter values for which the orbits of all critical points remains
bounded.

The Julia set for a given parameter `c` is the set of complex numbers for which
the function `f_c(z)` is bounded under iteration.

AUTHORS:

- Ben Barros
"""

#*****************************************************************************
#       Copyright (C) 2017 BEN BARROS <bbarros@slu.edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from sage.dynamics.complex_dynamics.mandel_julia_helper import (fast_mandelbrot_plot,
                                                                fast_external_ray,
                                                                convert_to_pixels,
                                                                get_line,
                                                                fast_julia_plot,
                                                                general_julia,
                                                                polynomial_mandelbrot,
                                                                julia_helper)

from sage.misc.lazy_import import lazy_import
lazy_import("sage.plot.colors", "Color")
from sage.repl.image import Image
from sage.functions.log import logb
from sage.functions.other import floor
from sage.rings.rational_field import QQ
from sage.rings.cc import CC
from sage.rings.complex_double import CDF
from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
from sage.schemes.projective.projective_space import ProjectiveSpace
from sage.misc.prandom import randint
from sage.calculus.var import var
from sage.rings.fraction_field import FractionField_generic
from sage.categories.function_fields import FunctionFields

lazy_import('sage.dynamics.arithmetic_dynamics.generic_ds', 'DynamicalSystem')

EPS = 0.00001


def mandelbrot_plot(f=None, **kwds):
    r"""
    Plot of the Mandelbrot set for a one parameter family of polynomial maps.

    The family `f_c(z)` must have parent ``R`` of the
    form ``R.<z,c> = CC[]``.

    REFERENCE:

    [Dev2005]_

    INPUT:

    - ``f`` -- map (default: ``z^2 + c``); polynomial family used to
      plot the Mandelbrot set

    - ``parameter`` -- variable (default: ``c``); parameter variable
      used to plot the Mandelbrot set

    - ``x_center`` -- double (default: ``-1.0``); Real part of center
      point

    - ``y_center`` -- double (default: ``0.0``); Imaginary part of
      center point

    - ``image_width`` -- double (default: ``4.0``); width of image
      in the complex plane

    - ``max_iteration`` -- long (default: ``500``); maximum number of
      iterations the map ``f_c(z)``

    - ``pixel_count`` -- long (default: ``500``); side length of
      image in number of pixels

    - ``base_color`` -- RGB color (default: ``[40, 40, 40]``); color
      used to determine the coloring of set

    - ``level_sep`` -- long (default: 1); number of iterations
      between each color level

    - ``number_of_colors`` -- long (default: 30); number of colors
      used to plot image

    - ``interact`` -- boolean (default: ``False``); controls whether
      plot will have interactive functionality

    OUTPUT: 24-bit RGB image of the Mandelbrot set in the complex plane

    EXAMPLES:

    ::

        sage: mandelbrot_plot()
        500x500px 24-bit RGB image

    ::

        sage: mandelbrot_plot(pixel_count=1000)
        1000x1000px 24-bit RGB image

    ::

        sage: mandelbrot_plot(x_center=-1.11, y_center=0.2283, image_width=1/128, # long time
        ....: max_iteration=2000, number_of_colors=500, base_color=[40, 100, 100])
        500x500px 24-bit RGB image

    To display an interactive plot of the Mandelbrot in the Notebook, set
    ``interact`` to ``True``. (This is only implemented for ``z^2 + c``)::

        sage: mandelbrot_plot(interact=True)
        ...interactive(children=(FloatSlider(value=0.0, description='Real center', max=1.0, min=-1.0, step=1e-05),
        FloatSlider(value=0.0, description='Imag center', max=1.0, min=-1.0, step=1e-05),
        FloatSlider(value=4.0, description='Width', max=4.0, min=1e-05, step=1e-05),
        IntSlider(value=500, description='Iterations', max=1000),
        IntSlider(value=500, description='Pixels', max=1000, min=10),
        IntSlider(value=1, description='Color sep', max=20, min=1),
        IntSlider(value=30, description='# Colors', min=1),
        ColorPicker(value='#ff6347', description='Base color'), Output()),
        _dom_classes=('widget-interact',))

    ::

        sage: mandelbrot_plot(interact=True, x_center=-0.75, y_center=0.25,
        ....: image_width=1/2, number_of_colors=75)
        ...interactive(children=(FloatSlider(value=-0.75, description='Real center', max=1.0, min=-1.0, step=1e-05),
        FloatSlider(value=0.25, description='Imag center', max=1.0, min=-1.0, step=1e-05),
        FloatSlider(value=0.5, description='Width', max=4.0, min=1e-05, step=1e-05),
        IntSlider(value=500, description='Iterations', max=1000),
        IntSlider(value=500, description='Pixels', max=1000, min=10),
        IntSlider(value=1, description='Color sep', max=20, min=1),
        IntSlider(value=75, description='# Colors', min=1),
        ColorPicker(value='#ff6347', description='Base color'), Output()),
        _dom_classes=('widget-interact',))

    Polynomial maps can be defined over a multivariate polynomial ring or a
    univariate polynomial ring tower::

        sage: R.<z,c> = CC[]
        sage: f = z^2 + c
        sage: mandelbrot_plot(f)
        500x500px 24-bit RGB image

    ::

        sage: B.<c> = CC[]
        sage: R.<z> = B[]
        sage: f = z^5 + c
        sage: mandelbrot_plot(f)  # long time
        500x500px 24-bit RGB image

    When the polynomial is defined over a multivariate polynomial ring it is
    necessary to specify the parameter variable (default parameter is ``c``)::

        sage: R.<a,b> = CC[]
        sage: f = a^2 + b^3
        sage: mandelbrot_plot(f, parameter=b)  # long time
        500x500px 24-bit RGB image

    Interact functionality is not implemented for general polynomial maps::

        sage: R.<z,c> = CC[]
        sage: f = z^3 + c
        sage: mandelbrot_plot(f, interact=True)
        Traceback (most recent call last):
        ...
        NotImplementedError: interact only implemented for z^2 + c
    """
    parameter = kwds.pop("parameter", None)
    x_center = kwds.pop("x_center", 0.0)
    y_center = kwds.pop("y_center", 0.0)
    image_width = kwds.pop("image_width", 4.0)
    max_iteration = kwds.pop("max_iteration", None)
    pixel_count = kwds.pop("pixel_count", 500)
    level_sep = kwds.pop("level_sep", 1)
    number_of_colors = kwds.pop("number_of_colors", 30)
    interacts = kwds.pop("interact", False)
    base_color = kwds.pop("base_color", Color('tomato'))
    # Check if user specified maximum number of iterations
    given_iterations = True
    if max_iteration is None:
        # Set default to 500 for z^2 + c map
        max_iteration = 500
        given_iterations = False

    from ipywidgets.widgets import FloatSlider, IntSlider, ColorPicker, interact
    widgets = dict(
                   x_center=FloatSlider(min=-1.0, max=1.0, step=EPS,
                                          value=x_center, description="Real center"),
                   y_center=FloatSlider(min=-1.0, max=1.0, step=EPS,
                                          value=y_center, description="Imag center"),
                   image_width=FloatSlider(min=EPS, max=4.0, step=EPS,
                                             value=image_width, description='Width'),
                   max_iteration=IntSlider(min=0, max=1000,
                                             value=max_iteration, description='Iterations'),
                   pixel_count=IntSlider(min=10, max=1000,
                                           value=pixel_count, description='Pixels'),
                   level_sep=IntSlider(min=1, max=20,
                                         value=level_sep, description="Color sep"),
                   color_num=IntSlider(min=1, max=100,
                                         value=number_of_colors, description="# Colors"),
                   base_color=ColorPicker(value=Color(base_color).html_color(),
                                            description="Base color"),
                   )

    if f is None:
        # Quadratic map f = z^2 + c

        if interacts:
            return interact(**widgets).widget(fast_mandelbrot_plot)

        return fast_mandelbrot_plot(x_center, y_center, image_width,
         max_iteration, pixel_count, level_sep, number_of_colors,
         base_color)

    if parameter is None:
        c = var('c')
        parameter = c

    P = f.parent()

    if P.base_ring() is CC or P.base_ring() is CDF:
        if isinstance(P, FractionField_generic):
            raise NotImplementedError("coefficients must be polynomials in the parameter")
        gen_list = list(P.gens())
        parameter = gen_list.pop(gen_list.index(parameter))
        variable = gen_list.pop()

    elif P.base_ring().base_ring() is CC or P.base_ring().base_ring() is CDF:
        if isinstance(P.base_ring(), FractionField_generic):
            raise NotImplementedError("coefficients must be polynomials in the parameter")
        phi = P.flattening_morphism()
        f = phi(f)
        gen_list = list(f.parent().gens())
        parameter = gen_list.pop(gen_list.index(parameter))
        variable = gen_list.pop()

    elif P.base_ring() in FunctionFields():
        raise NotImplementedError("coefficients must be polynomials in the parameter")

    else:
        raise ValueError("base ring must be a complex field")

    if f == variable**2 + parameter:
        # Quadratic map f = z^2 + c
        if interacts:
            return interact(**widgets).widget(fast_mandelbrot_plot)

        return fast_mandelbrot_plot(x_center, y_center, image_width,
         max_iteration, pixel_count, level_sep, number_of_colors,
         base_color)
    if interacts:
        raise NotImplementedError("interact only implemented for z^2 + c")
    else:
        # Set default of max_iteration to 50 for general polynomial maps
        # This prevents the function from being very slow by default
        if not given_iterations:
            max_iteration = 50

        # Mandelbrot of General Polynomial Map
        return polynomial_mandelbrot(f, parameter, x_center, y_center,
         image_width, max_iteration, pixel_count, level_sep,
         number_of_colors, base_color)


def external_ray(theta, **kwds):
    r"""
    Draws the external ray(s) of a given angle (or list of angles)
    by connecting a finite number of points that were approximated using
    Newton's method. The algorithm used is described in a paper by
    Tomoki Kawahira.

    REFERENCE:

    [Kaw2009]_

    INPUT:

    - ``theta`` -- double or list of doubles, angles between 0 and 1 inclusive

    kwds:

    - ``image`` -- 24-bit RGB image (default: ``None``); user specified
      image of Mandelbrot set

    - ``D`` -- long (default: ``25``); depth of the approximation.
      As ``D`` increases, the external ray gets closer to the boundary of the
      Mandelbrot set. If the ray doesn't reach the boundary of the Mandelbrot
      set, increase ``D``.

    - ``S`` -- long (default: ``10``); sharpness of the approximation.
      Adjusts the number of points used to approximate the external ray (number
      of points is equal to ``S*D``). If ray looks jagged, increase ``S``.

    - ``R`` -- long (default: ``100``); radial parameter. If ``R`` is
      large, the external ray reaches sufficiently close to infinity. If ``R`` is
      too small, Newton's method may not converge to the correct ray.

    - ``prec`` -- long (default: ``300``); specifies the bits of
      precision used by the Complex Field when using Newton's method to compute
      points on the external ray

    - ``ray_color`` -- RGB color (default: ``[255, 255, 255]``) color
      of the external ray(s)

    OUTPUT: 24-bit RGB image of external ray(s) on the Mandelbrot set

    EXAMPLES::

        sage: external_ray(1/3)
        500x500px 24-bit RGB image

    ::

        sage: external_ray(0.6, ray_color=[255, 0, 0])
        500x500px 24-bit RGB image

    ::

        sage: external_ray([0, 0.2, 0.4, 0.7])
        500x500px 24-bit RGB image

    ::

        sage: external_ray([i/5 for i in range(1,5)])
        500x500px 24-bit RGB image

    WARNING:

    If you are passing in an image, make sure you specify
    which parameters to use when drawing the external ray.
    For example, the following is incorrect::

        sage: M = mandelbrot_plot(x_center=0)  # not tested
        sage: external_ray(5/7, image=M)       # not tested
        500x500px 24-bit RGB image

    To get the correct external ray, we adjust our parameters::

        sage: M = mandelbrot_plot(x_center=0)
        sage: external_ray(5/7, x_center=0, image=M)
        500x500px 24-bit RGB image

    .. TODO::

        The ``copy()`` function for bitmap images needs to be implemented
        in Sage.
    """
    x_0 = kwds.get("x_center", -1)
    y_0 = kwds.get("y_center", 0)
    plot_width = kwds.get("image_width", 4)
    pixel_width = kwds.get("pixel_count", 500)
    depth = kwds.get("D", 25)
    sharpness = kwds.get("S", 10)
    radial_parameter = kwds.get("R", 100)
    precision = kwds.get("prec", 300)
    precision = max(precision, -logb(pixel_width * 0.001, 2).round() + 10)
    ray_color = kwds.get("ray_color", [255] * 3)
    image = kwds.get("image", None)
    if image is None:
        image = mandelbrot_plot(x_center=x_0, **kwds)

    # Make a copy of the bitmap image.
    old_pixel = image.pixels()
    M = Image('RGB', (pixel_width, pixel_width))
    pixel = M.pixels()
    for i in range(pixel_width):
        for j in range(pixel_width):
            pixel[i, j] = old_pixel[i, j]

    # Make sure that theta is a list so loop below works
    if not isinstance(theta, list):
        theta = [theta]

    # Check if theta is in the interval [0,1]
    for angle in theta:
        if angle < 0 or angle > 1:
            raise ValueError("values for theta must be in "
                             "the closed interval [0,1].")

    # Loop through each value for theta in list and plot the external ray.
    for angle in theta:
        E = fast_external_ray(angle, D=depth, S=sharpness, R=radial_parameter,
                              prec=precision, image_width=plot_width,
                              pixel_count=pixel_width)

        # Convert points to pixel coordinates.
        pixel_list = convert_to_pixels(E, x_0, y_0, plot_width, pixel_width)

        # Find the pixels between points in pixel_list.
        extra_points = []
        for i in range(len(pixel_list) - 1):
            if min(pixel_list[i + 1]) >= 0 and max(pixel_list[i + 1]) < pixel_width:
                for j in get_line(pixel_list[i], pixel_list[i + 1]):
                    extra_points.append(j)

        # Add these points to pixel_list to fill in gaps in the ray.
        pixel_list += extra_points

        # Remove duplicates from list.
        pixel_list = list(set(pixel_list))

        # Check if point is in window and if it is, plot it on the image to
        # create an external ray.
        for k in pixel_list:
            if max(k) < pixel_width and min(k) >= 0:
                pixel[int(k[0]), int(k[1])] = tuple(ray_color)
    return M


def kneading_sequence(theta):
    r"""
    Determines the kneading sequence for an angle theta in RR/ZZ which
    is periodic under doubling. We use the definition for the kneading
    sequence given in Definition 3.2 of [LS1994]_.

    INPUT:

    - ``theta`` -- a rational number with odd denominator

    OUTPUT: string representing the kneading sequence of theta in RR/ZZ

    REFERENCES:

    [LS1994]_

    EXAMPLES::

        sage: kneading_sequence(0)
        '*'

    ::

        sage: kneading_sequence(1/3)
        '1*'

    Since 1/3 and 7/3 are the same in RR/ZZ, they have the same kneading sequence::

        sage: kneading_sequence(7/3)
        '1*'

    We can also use (finite) decimal inputs, as long as the denominator in reduced form is odd::

        sage: kneading_sequence(1.2)
        '110*'

    Since rationals with even denominator are not periodic under doubling, we have not implemented kneading sequences for such rationals::

        sage: kneading_sequence(1/4)
        Traceback (most recent call last):
        ...
        ValueError: input must be a rational number with odd denominator
    """

    if theta not in QQ:
        raise TypeError('input must be a rational number with odd denominator')
    elif QQ(theta).valuation(2) < 0:
        raise ValueError('input must be a rational number with odd denominator')
    else:
        theta = QQ(theta)
        theta = theta - floor(theta)
        KS = []
        not_done = True
        left = theta/2
        right = (theta + 1)/2
        y = theta
        while not_done:
            if ((y < left) or (y > right)):
                KS.append('0')
            elif ((y > left) and (y < right)):
                KS.append('1')
            else:
                not_done = False
            y = 2*y - floor(2*y)
        KS_str = ''.join(KS) + '*'
    return KS_str


def julia_plot(f=None, **kwds):
    r"""
    Plots the Julia set of a given polynomial ``f``. Users can specify whether
    they would like to display the Mandelbrot side by side with the Julia set
    with the ``mandelbrot`` argument. If ``f`` is not specified, this method
    defaults to `f(z) = z^2-1`.

    The Julia set of a polynomial ``f`` is the set of complex numbers `z` for
    which the function `f(z)` is bounded under iteration. The Julia set can
    be visualized by plotting each point in the set in the complex plane.
    Julia sets are examples of fractals when plotted in the complex plane.

    ALGORITHM:

    Let `R_c = \bigl(1 + \sqrt{1 + 4|c|}\bigr)/2` if the polynomial is of the
    form `f(z) = z^2 + c`; otherwise, let `R_c = 2`.
    For every `p \in \mathbb{C}`, if `|f^{k}(p)| > R_c` for some `k \geq 0`,
    then `f^{n}(p) \to \infty`.  Let `N` be the maximum number of iterations.
    Compute the first `N` points on the orbit of `p` under `f`. If for
    any `k < N`, `|f^{k}(p)| > R_c`, we stop the iteration and assign a color
    to the point `p` based on how quickly `p` escaped to infinity under
    iteration of `f`. If `|f^{i}(p)| \leq R_c` for all `i \leq N`, we assume
    `p` is in the Julia set and assign the point `p` the color black.

    INPUT:

    - ``f`` -- input polynomial (default: ``z^2 - 1``)

    - ``period`` -- list (default: ``None``); returns the Julia set
      for a random `c` value with the given (formal) cycle structure

    - ``mandelbrot`` -- boolean (default: ``True``); when set to
      ``True``, an image of the Mandelbrot set is appended to the right of the
      Julia set

    - ``point_color`` -- RGB color (default: ``'tomato'``);
      color of the point `c` in the Mandelbrot set (any valid input for Color)

    - ``x_center`` -- double (default: ``-1.0``); real part
      of center point

    - ``y_center`` -- double (default: ``0.0``); imaginary part
      of center point

    - ``image_width`` -- double (default: ``4.0``); width of image
      in the complex plane

    - ``max_iteration`` -- long (default: ``500``); maximum number
      of iterations the map `f(z)`

    - ``pixel_count`` -- long (default: ``500``); side length of
      image in number of pixels

    - ``base_color`` -- hex color (default: ``'steelblue'``); color
      used to determine the coloring of set (any valid input for Color)

    - ``level_sep`` -- long (default: 1); number of iterations
      between each color level

    - ``number_of_colors`` -- long (default: 30); number of colors
      used to plot image

    - ``interact`` -- boolean (default: ``False``); controls whether
      plot will have interactive functionality

    OUTPUT: 24-bit RGB image of the Julia set in the complex plane

    .. TODO::

        Implement the side-by-side Mandelbrot-Julia plots for general one-parameter families
        of polynomials.

    EXAMPLES:

    The default ``f`` is `z^2 - 1`::

        sage: julia_plot()
        1001x500px 24-bit RGB image

    To display only the Julia set, set ``mandelbrot`` to ``False``::

        sage: julia_plot(mandelbrot=False)
        500x500px 24-bit RGB image

    ::

        sage: R.<z> = CC[]
        sage: f = z^3 - z + 1
        sage: julia_plot(f)  # long time
        500x500px 24-bit RGB image

    To display an interactive plot of the Julia set in the Notebook,
    set ``interact`` to ``True``. (This is only implemented for polynomials of
    the form ``f = z^2 + c``)::

        sage: julia_plot(interact=True)
        ...interactive(children=(FloatSlider(value=-1.0, description='Real c'...

        ::

        sage: R.<z> = CC[]
        sage: f = z^2 + 1/2
        sage: julia_plot(f,interact=True)
        ...interactive(children=(FloatSlider(value=0.5, description='Real c'...

    To return the Julia set of a random `c` value with (formal) cycle structure
    `(2,3)`, set ``period = [2,3]``::

        sage: julia_plot(period=[2,3])
        1001x500px 24-bit RGB image

    To return all of the Julia sets of `c` values with (formal) cycle structure
    `(2,3)`::

        sage: period = [2,3] # not tested
        ....: R.<c> = QQ[]
        ....: P.<x,y> = ProjectiveSpace(R,1)
        ....: f = DynamicalSystem([x^2+c*y^2, y^2])
        ....: L = f.dynatomic_polynomial(period).subs({x:0,y:1}).roots(ring=CC)
        ....: c_values = [k[0] for k in L]
        ....: for c in c_values:
        ....:     julia_plot(c)

    Polynomial maps can be defined over a polynomial ring or a fraction field,
    so long as ``f`` is polynomial::

        sage: R.<z> = CC[]
        sage: f = z^2 - 1
        sage: julia_plot(f)
        1001x500px 24-bit RGB image

    ::

        sage: R.<z> = CC[]
        sage: K = R.fraction_field(); z = K.gen()
        sage: f = z^2-1
        sage: julia_plot(f)
        1001x500px 24-bit RGB image

    Interact functionality is not implemented if the polynomial is not of the
    form `f = z^2 + c`::

        sage: R.<z> = CC[]
        sage: f = z^3 + 1
        sage: julia_plot(f, interact=True)
        Traceback (most recent call last):
        ...
        NotImplementedError: The interactive plot is only implemented for ...
    """

    # extract keyword arguments
    period = kwds.pop("period", None)
    mandelbrot = kwds.pop("mandelbrot", True)
    point_color = kwds.pop("point_color", 'tomato')
    x_center = kwds.pop("x_center", 0.0)
    y_center = kwds.pop("y_center", 0.0)
    image_width = kwds.pop("image_width", 4.0)
    max_iteration = kwds.pop("max_iteration", 500)
    pixel_count = kwds.pop("pixel_count", 500)
    base_color = kwds.pop("base_color", 'steelblue')
    level_sep = kwds.pop("level_sep", 1)
    number_of_colors = kwds.pop("number_of_colors", 30)
    interacts = kwds.pop("interact", False)

    f_is_default_after_all = None

    if period: # pick a random c with the specified period
        R = PolynomialRing(CC, 'c')
        c = R.gen()
        x, y = ProjectiveSpace(R, 1, 'x,y').gens()
        F = DynamicalSystem([x**2 + c * y**2, y**2])
        L = F.dynatomic_polynomial(period).subs({x: 0, y: 1}).roots(ring=CC)
        c = L[randint(0, len(L) - 1)][0]

    base_color = Color(base_color)
    point_color = Color(point_color)

    EPS = 0.00001

    if f is not None and period is None: # f user-specified and no period given

        # try to coerce f to live in a polynomial ring
        S = PolynomialRing(CC, names='z')
        z = S.gen()
        try:
            f_poly = S(f)
        except TypeError:
            R = f.parent()
            if not (R.is_integral_domain() and
                   (CC.is_subring(R) or CDF.is_subring(R))):
                raise ValueError('given `f` must be a complex polynomial')
            raise NotImplementedError(
                'Julia sets not implemented for rational functions')

        if (f_poly - z*z) in CC: # f is specified and of the form z^2 + c.
            f_is_default_after_all = True
            c = f_poly - z*z
        else: # f is specified and not of the form z^2 + c
            if interacts:
                raise NotImplementedError(
                        "The interactive plot is only implemented for "
                        "polynomials of the form f = z^2 + c."
                            )
            else:
                return general_julia(f_poly, x_center, y_center,
                                    image_width, max_iteration,
                                    pixel_count, level_sep,
                                    number_of_colors, base_color)

    # otherwise we can use fast_julia_plot for z^2 + c
    if f_is_default_after_all or f is None or period is not None:

        # specify default c = -1 value if f and period were not specified
        if not f_is_default_after_all and period is None:
            c = -1

        c = CC(c)
        c_real = c.real()
        c_imag = c.imag()

        if interacts: # set widgets
            from ipywidgets.widgets import FloatSlider, IntSlider, \
                                           ColorPicker, interact
            widgets = dict(
                c_real=FloatSlider(min=-2.0, max=2.0, step=EPS,
                                  value=c_real, description="Real c"),
                c_imag=FloatSlider(min=-2.0, max=2.0, step=EPS,
                                  value=c_imag, description="Imag c"),
                x_center=FloatSlider(min=-1.0, max=1.0, step=EPS,
                                  value=x_center, description="Real center"),
                y_center=FloatSlider(min=-1.0, max=1.0, step=EPS,
                                  value=y_center, description="Imag center"),
                image_width=FloatSlider(min=EPS, max=4.0, step=EPS,
                                  value=image_width, description='Width'),
                max_iteration=IntSlider(min=0, max=1000,
                                  value=max_iteration, description='Iterations'),
                pixel_count=IntSlider(min=10, max=1000,
                                  value=pixel_count, description='Pixels'),
                level_sep=IntSlider(min=1, max=20,
                                  value=level_sep, description="Color sep"),
                color_num=IntSlider(min=1, max=100,
                                  value=number_of_colors, description="# Colors"),
                base_color=ColorPicker(value=base_color.html_color(),
                                         description="Base color"),
            )
            if mandelbrot:
                widgets["point_color"] = ColorPicker(value=point_color.html_color(),
                                                     description="Point color")
                return interact(**widgets).widget(julia_helper)
            return interact(**widgets).widget(fast_julia_plot)
        if mandelbrot:  # non-interactive with mandelbrot
            return julia_helper(c_real, c_imag, x_center, y_center,
                                image_width, max_iteration, pixel_count,
                                level_sep, number_of_colors, base_color,
                                point_color)
        # non-interactive without mandelbrot
        return fast_julia_plot(c_real, c_imag, x_center, y_center,
                               image_width, max_iteration, pixel_count,
                               level_sep, number_of_colors, base_color)



from sage.libs.flint.types cimport fmpz_t, fmpq_mat_t
from sage.matrix.matrix_dense cimport Matrix_dense

cdef class Matrix_rational_dense(Matrix_dense):
    cdef fmpq_mat_t _matrix

    cdef int fmpz_denom(self, fmpz_t d) except -1
    cdef int fmpz_height(self, fmpz_t height) except -1
#    cdef int _rescale(self, mpq_t a) except -1

    cdef _pickle_version0(self)
    cdef _unpickle_version0(self, data)
    cpdef _export_as_string(self, int base=?)

    cdef _add_ui_unsafe_assuming_int(self, Py_ssize_t i, Py_ssize_t j, unsigned long int n)
    cdef _sub_ui_unsafe_assuming_int(self, Py_ssize_t i, Py_ssize_t j, unsigned long int n)

    cdef inline Matrix_rational_dense _new_matrix(self, Py_ssize_t nrows, Py_ssize_t ncols)

cdef class MatrixWindow:
    cdef Matrix_rational_dense _matrix
    cdef int _row, _col, _nrows, _ncols



r"""
`p`-adic Valuations on Number Fields and Their Subrings and Completions

EXAMPLES::

    sage: ZZ.valuation(2)
    2-adic valuation
    sage: QQ.valuation(3)
    3-adic valuation
    sage: CyclotomicField(5).valuation(5)                                               # needs sage.rings.number_field
    5-adic valuation
    sage: GaussianIntegers().valuation(7)                                               # needs sage.rings.number_field
    7-adic valuation
    sage: Zp(11).valuation()
    11-adic valuation

These valuations can then, e.g., be used to compute approximate factorizations
in the completion of a ring::

    sage: v = ZZ.valuation(2)
    sage: R.<x> = ZZ[]
    sage: f = x^5 + x^4 + x^3 + x^2 + x - 1
    sage: v.montes_factorization(f, required_precision=20)                              # needs sage.geometry.polyhedron
    (x + 676027) * (x^4 + 372550*x^3 + 464863*x^2 + 385052*x + 297869)

They can also be called on elements, although it is usually shorter to use ``.valuation()``::

    sage: v(24)
    3
    sage: 24.valuation(2)
    3
    sage: 24.valuation(v.uniformizer())
    3

AUTHORS:

- Julian Rüth (2013-03-16): initial version

REFERENCES:

The theory used here was originally developed in [Mac1936I]_ and [Mac1936II]_. An
overview can also be found in Chapter 4 of [Rüt2014]_.
"""
#*****************************************************************************
#       Copyright (C) 2013-2020 Julian Rüth <julian.rueth@fsfe.org>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************
from sage.rings.valuation.valuation import DiscreteValuation
from sage.rings.valuation.value_group import DiscreteValueSemigroup
from sage.rings.valuation.mapped_valuation import FiniteExtensionFromLimitValuation
from sage.structure.factory import UniqueFactory
from sage.misc.cachefunc import cached_method

from sage.rings.infinity import infinity


class PadicValuationFactory(UniqueFactory):
    r"""
    Create a ``prime``-adic valuation on ``R``.

    INPUT:

    - ``R`` -- a subring of a number field or a subring of a local field in
      characteristic zero

    - ``prime`` -- a prime that does not split, a discrete (pseudo-)valuation,
      a fractional ideal, or ``None`` (default: ``None``)

    EXAMPLES:

    For integers and rational numbers, ``prime`` is just a prime of the
    integers::

        sage: valuations.pAdicValuation(ZZ, 3)
        3-adic valuation

        sage: valuations.pAdicValuation(QQ, 3)
        3-adic valuation

    ``prime`` may be ``None`` for local rings::

        sage: valuations.pAdicValuation(Qp(2))
        2-adic valuation

        sage: valuations.pAdicValuation(Zp(2))
        2-adic valuation

    But it must be specified in all other cases::

        sage: valuations.pAdicValuation(ZZ)
        Traceback (most recent call last):
        ...
        ValueError: prime must be specified for this ring

    It can sometimes be beneficial to define a number field extension as a
    quotient of a polynomial ring (since number field extensions always compute
    an absolute polynomial defining the extension which can be very costly)::

        sage: R.<x> = QQ[]
        sage: K.<a> = NumberField(x^2 + 1)
        sage: R.<x> = K[]
        sage: L.<b> = R.quo(x^2 + a)
        sage: valuations.pAdicValuation(L, 2)
        2-adic valuation

    .. SEEALSO::

        :meth:`NumberField_generic.valuation() <sage.rings.number_field.number_field.NumberField_generic.valuation>`,
        :meth:`Order.valuation() <sage.rings.number_field.order.Order.valuation>`,
        :meth:`pAdicGeneric.valuation() <sage.rings.padics.padic_generic.pAdicGeneric.valuation>`,
        :meth:`RationalField.valuation() <sage.rings.rational_field.RationalField.valuation>`,
        :meth:`IntegerRing_class.valuation() <sage.rings.integer_ring.IntegerRing_class.valuation>`.
    """
    def create_key_and_extra_args(self, R, prime=None, approximants=None):
        r"""
        Create a unique key identifying the valuation of ``R`` with respect to
        ``prime``.

        EXAMPLES::

            sage: QQ.valuation(2)  # indirect doctest
            2-adic valuation
        """
        from sage.rings.integer_ring import ZZ
        from sage.rings.rational_field import QQ
        from sage.rings.padics.padic_generic import pAdicGeneric
        from sage.rings.number_field.number_field_base import NumberField
        from sage.rings.polynomial.polynomial_quotient_ring import PolynomialQuotientRing_generic

        if R.characteristic() != 0:
            # We do not support equal characteristic yet
            raise ValueError("R must be a ring of characteristic zero.")

        if R is ZZ or R is QQ:
            return self.create_key_for_integers(R, prime), {}
        if isinstance(R, pAdicGeneric):
            return self.create_key_for_local_ring(R, prime), {}
        if isinstance(R.fraction_field(), NumberField) or isinstance(R, PolynomialQuotientRing_generic):
            return self.create_key_and_extra_args_for_number_field(R, prime, approximants=approximants)
        raise NotImplementedError("p-adic valuations not implemented for %r" % (R,))

    def create_key_for_integers(self, R, prime):
        r"""
        Create a unique key identifying the valuation of ``R`` with respect to
        ``prime``.

        EXAMPLES::

            sage: QQ.valuation(2)  # indirect doctest
            2-adic valuation
        """
        from sage.rings.integer_ring import ZZ
        if prime is None:
            raise ValueError("prime must be specified for this ring")
        from sage.rings.valuation.valuation import DiscretePseudoValuation
        if isinstance(prime, DiscretePseudoValuation):
            prime = prime.uniformizer()
        if prime not in ZZ or not ZZ(prime).is_prime():
            raise ValueError("prime must be a prime in the integers but %s is not" % (prime,))
        return R, prime

    def create_key_for_local_ring(self, R, prime):
        r"""
        Create a unique key identifying the valuation of ``R`` with respect to
        ``prime``.

        EXAMPLES::

            sage: Qp(2).valuation()  # indirect doctest
            2-adic valuation
        """
        # We do not care much about the value of prime since there is only one
        # reasonable p-adic valuation here
        if prime is not None:
            if prime in R:
                if R(prime).valuation() <= 0:
                    raise ValueError("prime must be an element of positive valuation")
            elif prime(R.prime()) <= 0:
                raise ValueError("prime must be an element of positive valuation")

        return (R,)

    def create_key_and_extra_args_for_number_field(self, R, prime, approximants):
        r"""
        Create a unique key identifying the valuation of ``R`` with respect to
        ``prime``.

        EXAMPLES::

            sage: GaussianIntegers().valuation(2)  # indirect doctest                   # needs sage.rings.number_field
            2-adic valuation
        """
        K, L, G = self._normalize_number_field_data(R)

        from sage.rings.number_field.number_field_ideal import NumberFieldFractionalIdeal
        from sage.rings.valuation.valuation import DiscretePseudoValuation
        if isinstance(prime, DiscretePseudoValuation):
            return self.create_key_and_extra_args_for_number_field_from_valuation(R, prime, prime, approximants=approximants)
        if prime in K:
            return self.create_key_and_extra_args_for_number_field_from_valuation(R, K.valuation(prime), prime, approximants=approximants)
        if prime in L or isinstance(prime, NumberFieldFractionalIdeal):
            return self.create_key_and_extra_args_for_number_field_from_ideal(R, L.fractional_ideal(prime), prime)
        raise ValueError("prime must be a discrete pseudo-valuation, a prime in the base ring, or a fractional ideal")

    def create_key_and_extra_args_for_number_field_from_valuation(self, R, v, prime, approximants):
        r"""
        Create a unique key identifying the valuation of ``R`` with respect to
        ``v``.

        .. NOTE::

            ``prime``, the original parameter that was passed to
            :meth:`create_key_and_extra_args`, is only used to provide more
            meaningful error messages

        EXAMPLES::

            sage: GaussianIntegers().valuation(ZZ.valuation(2))  # indirect doctest     # needs sage.rings.number_field
            2-adic valuation

        TESTS:

        We can extend to the field of fractions of a quotient ring::

            sage: R.<x> = ZZ[]
            sage: S = R.quo(x^2 + 1)
            sage: v = valuations.pAdicValuation(S, 2)                                   # needs sage.geometry.polyhedron
            sage: R.<x> = QQ[]
            sage: S = R.quo(x^2 + 1)
            sage: v = valuations.pAdicValuation(S, v)                                   # needs sage.geometry.polyhedron
        """
        K, L, G = self._normalize_number_field_data(R)

        if v.domain().is_subring(G.parent()):
            # v is defined on a subring of K[x].
            # We try to lift v to a pseudo-valuation on K[x].
            if _fraction_field(v.domain()) is not _fraction_field(G.parent()):
                # First, we lift valuations defined on subrings of K to
                # valuations on K[x].
                if v.domain().is_subring(K):
                    if v.domain() is not K:
                        v = K.valuation(v)
                    from sage.rings.valuation.gauss_valuation import GaussValuation
                    v = GaussValuation(G.parent(), v)
            if v.domain() != G.parent():
                # Then, we lift valuations defined on polynomial rings which are
                # subrings of K[x] to K[x]
                v = v.extension(G.parent())
        elif _fraction_field(v.domain()) == L:
            # v is defined on a ring whose field of fractions is L
            v = v._base_valuation._initial_approximation.change_domain(G.parent())
        else:
            raise NotImplementedError("cannot rewrite %r which is defined on %r as a pseudo-valuation on %r" % (v, v.domain(), G.parent()))

        assert (v.domain() is G.parent())

        # To obtain uniqueness of p-adic valuations, we need a canonical
        # description of v. We consider all extensions of vK to L and select
        # the one approximated by v.
        vK = v.restriction(v.domain().base_ring()).extension(K)
        if approximants is None:
            approximants = vK.mac_lane_approximants(G, require_incomparability=True)
        approximants = [approximant.extension(v.domain()) for approximant in approximants]
        approximant = vK.mac_lane_approximant(G, v, approximants=tuple(approximants))

        return (R, approximant), {'approximants': approximants}

    def create_key_and_extra_args_for_number_field_from_ideal(self, R, I, prime):
        r"""
        Create a unique key identifying the valuation of ``R`` with respect to
        ``I``.

        .. NOTE::

            ``prime``, the original parameter that was passed to
            :meth:`create_key_and_extra_args`, is only used to provide more
            meaningful error messages

        EXAMPLES::

            sage: GaussianIntegers().valuation(GaussianIntegers().number_field().fractional_ideal(2))  # indirect doctest
            2-adic valuation

        TESTS:

        Verify that :issue:`28976` has been resolved::

            sage: R.<x> = QQ[]
            sage: K.<a> = NumberField(x^6 - 18*x^4 - 24*x^3 + 27*x^2 + 36*x - 6)
            sage: I = K.fractional_ideal((2, -7/44*a^5 + 19/44*a^4 + 87/44*a^3 - 87/44*a^2 - 5/2*a + 39/22))
            sage: I.norm()
            2
            sage: I in K.primes_above(2)
            True
            sage: K.valuation(I)
            [ 2-adic valuation, v(x + 1) = 1/2 ]-adic valuation

        ::

            sage: K.<a, b> = NumberField([x^2 - 2, x^2 + x + 1])
            sage: K.valuation(2)
            2-adic valuation
        """
        K, L, G = self._normalize_number_field_data(R)

        # To obtain uniqueness of p-adic valuations, we need a canonical
        # description of v. We consider all extensions of vK to L and select
        # the one approximated by v.
        # Of course, this only works if I comes from a single prime downstairs.
        p = I.relative_norm()
        F = p.factor()
        if len(F) != 1:
            raise ValueError("%r does not lie over a single prime of %r" % (I, K))
        vK = K.valuation(F[0][0])
        approximants = vK.mac_lane_approximants(G, require_incomparability=True)

        candidates = approximants[:]

        # The correct approximant has v(g) > 0 for all g in the ideal.
        # Unfortunately, the generators of I, even though defined over K have
        # their polynomial() defined over the rationals so we need to turn them
        # into polynomials over K[x] explicitly.
        from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
        gens = I.gens()
        gens = [PolynomialRing(K, 'x')(list(g.vector())) for g in gens]

        # Refine candidates until we can detect which valuation corresponds to the ideal I
        while True:
            assert any(candidates), "the defining polynomial of the extension factored but we still could not figure out which valuation corresponds to the given ideal"

            match = [i for (i, v) in enumerate(candidates) if v and all(v(g) > 0 for g in gens)]

            if len(match) > 1:
                raise ValueError("%s does not single out a unique extension of %s to %s" % (prime, vK, L))
            if len(match) == 1:
                return (R, approximants[match[0]]), {'approximants': approximants}

            # We refine candidates which increases v(g) for all g in I;
            # however, we cannot augment the valuations which are already at
            # v(G) = +∞ which we ignore by setting them to None.
            candidates = [v.mac_lane_step(G)[0] if v and v.is_discrete_valuation() else None for v in candidates]

    def _normalize_number_field_data(self, R):
        r"""
        Helper method which returns the defining data of the number field
        ``R``.

        EXAMPLES::

            sage: R.<x> = QQ[]
            sage: K = R.quo(x^2 + 1)
            sage: valuations.pAdicValuation._normalize_number_field_data(K)
            (Rational Field,
             Univariate Quotient Polynomial Ring in xbar over Rational Field with modulus x^2 + 1,
             x^2 + 1)
        """
        from sage.rings.polynomial.polynomial_quotient_ring import PolynomialQuotientRing_generic
        from sage.rings.number_field.number_field_base import NumberField
        if isinstance(R.fraction_field(), NumberField):
            L = R.fraction_field()
            G = L.relative_polynomial()
            K = L.base_ring()
        elif isinstance(R, PolynomialQuotientRing_generic):
            from sage.categories.number_fields import NumberFields
            if R.base_ring().fraction_field() not in NumberFields():
                raise NotImplementedError("cannot normalize quotients over %r" % (R.base_ring(),))
            L = R.fraction_field()
            K = R.base_ring().fraction_field()
            G = R.modulus().change_ring(K)
        else:
            raise NotImplementedError("cannot normalize %r" % (R,))

        return K, L, G

    def create_object(self, version, key, **extra_args):
        r"""
        Create a `p`-adic valuation from ``key``.

        EXAMPLES::

            sage: ZZ.valuation(5)  # indirect doctest
            5-adic valuation
        """
        from sage.rings.integer_ring import ZZ
        from sage.rings.rational_field import QQ
        from sage.rings.padics.padic_generic import pAdicGeneric
        from sage.rings.valuation.valuation_space import DiscretePseudoValuationSpace
        from sage.rings.polynomial.polynomial_quotient_ring import PolynomialQuotientRing_generic
        from sage.rings.number_field.number_field_base import NumberField
        R = key[0]
        parent = DiscretePseudoValuationSpace(R)
        if isinstance(R, pAdicGeneric):
            assert (len(key) == 1)
            return parent.__make_element_class__(pAdicValuation_padic)(parent)
        if R is ZZ or R is QQ:
            prime = key[1]
            assert (len(key) == 2)
            return parent.__make_element_class__(pAdicValuation_int)(parent, prime)
        v = key[1]
        approximants = extra_args['approximants']
        parent = DiscretePseudoValuationSpace(R)
        K = R.fraction_field()
        if isinstance(K, NumberField):
            G = K.relative_polynomial()
        elif isinstance(R, PolynomialQuotientRing_generic):
            G = R.modulus()
        else:
            raise NotImplementedError
        return parent.__make_element_class__(pAdicFromLimitValuation)(parent, v, G.change_ring(R.base_ring()), approximants)


pAdicValuation = PadicValuationFactory("sage.rings.padics.padic_valuation.pAdicValuation")


class pAdicValuation_base(DiscreteValuation):
    r"""
    Abstract base class for `p`-adic valuations.

    INPUT:

    - ``ring`` -- an integral domain

    - ``p`` -- a rational prime over which this valuation lies, not
      necessarily a uniformizer for the valuation

    EXAMPLES::

        sage: ZZ.valuation(3)
        3-adic valuation

        sage: QQ.valuation(5)
        5-adic valuation

     For `p`-adic rings, ``p`` has to match the `p` of the ring. ::

        sage: v = valuations.pAdicValuation(Zp(3), 2); v
        Traceback (most recent call last):
        ...
        ValueError: prime must be an element of positive valuation

    TESTS::

        sage: TestSuite(ZZ.valuation(3)).run()  # long time                             # needs sage.geometry.polyhedron
        sage: TestSuite(QQ.valuation(5)).run()  # long time                             # needs sage.geometry.polyhedron
        sage: TestSuite(Zp(5).valuation()).run()        # long time                     # needs sage.geometry.polyhedron
    """
    def __init__(self, parent, p):
        r"""
        TESTS::

            sage: from sage.rings.padics.padic_valuation import pAdicValuation_base
            sage: isinstance(ZZ.valuation(3), pAdicValuation_base)
            True
        """
        DiscreteValuation.__init__(self, parent)

        from sage.rings.integer_ring import ZZ
        self._p = ZZ(p)

    def p(self):
        r"""
        Return the `p` of this `p`-adic valuation.

        EXAMPLES::

            sage: GaussianIntegers().valuation(2).p()                                   # needs sage.rings.number_field
            2
        """
        return self._p

    def reduce(self, x):
        r"""
        Reduce ``x`` modulo the ideal of elements of positive valuation.

        INPUT:

        - ``x`` -- an element in the domain of this valuation

        OUTPUT: an element of the :meth:`~sage.rings.valuation.valuation_space.DiscretePseudoValuationSpace.ElementMethods.residue_field`

        EXAMPLES::

            sage: v = ZZ.valuation(3)
            sage: v.reduce(4)
            1
        """
        x = self.domain().coerce(x)

        if self(x) < 0:
            raise ValueError("reduction is only defined for elements of nonnegative valuation")

        return self.residue_field()(x)

    def lift(self, x):
        r"""
        Lift ``x`` from the residue field to the domain of this valuation.

        INPUT:

        - ``x`` -- an element of the :meth:`~sage.rings.valuation.valuation_space.DiscretePseudoValuationSpace.ElementMethods.residue_field`

        EXAMPLES::

            sage: v = ZZ.valuation(3)
            sage: xbar = v.reduce(4)
            sage: v.lift(xbar)
            1
        """
        x = self.residue_field().coerce(x)

        return self.domain()(x)

    def is_unramified(self, G, include_steps=False, assume_squarefree=False):
        r"""
        Return whether ``G`` defines a single unramified extension of the
        completion of the domain of this valuation.

        INPUT:

        - ``G`` -- a monic squarefree polynomial over the domain of this valuation

        - ``include_steps`` -- boolean (default: ``False``); whether to
          include the approximate valuations that were used to determine the
          result in the return value

        - ``assume_squarefree`` -- boolean (default: ``False``); whether to
          assume that ``G`` is square-free over the completion of the domain of
          this valuation. Setting this to ``True`` can significantly improve
          the performance.

        EXAMPLES:

        We consider an extension as unramified if its ramification index is 1.
        Hence, a trivial extension is unramified::

            sage: R.<x> = QQ[]
            sage: v = QQ.valuation(2)
            sage: v.is_unramified(x)
            True

        If ``G`` remains irreducible in reduction, then it defines an
        unramified extension::

            sage: v.is_unramified(x^2 + x + 1)
            True

        However, even if ``G`` factors, it might define an unramified
        extension::

            sage: v.is_unramified(x^2 + 2*x + 4)                                        # needs sage.geometry.polyhedron
            True
        """
        R = G.parent()

        from sage.rings.polynomial.polynomial_ring import PolynomialRing_generic
        if not isinstance(R, PolynomialRing_generic) or R.base_ring() is not self.domain() or not G.is_monic():
            raise ValueError("G must be a monic univariate polynomial over the domain of this valuation")
        if not assume_squarefree and not G.is_squarefree():
            raise ValueError("G must be squarefree")

        from sage.rings.valuation.gauss_valuation import GaussValuation

        steps = [ GaussValuation(R, self) ]
        while True:
            v = steps[-1]
            if v.E() > 1:
                ret = False
                break
            if v.F() == G.degree():
                ret = True
                break

            assert v(G) is not infinity
            if v.is_key(G):
                ret = True
                break

            next = v.mac_lane_step(G, assume_squarefree=True)
            if len(next) > 1:
                ret = False
                break
            steps.append(next[0])

        if include_steps:
            return ret, steps
        return ret

    def is_totally_ramified(self, G, include_steps=False, assume_squarefree=False):
        r"""
        Return whether ``G`` defines a single totally ramified extension of the
        completion of the domain of this valuation.

        INPUT:

        - ``G`` -- a monic squarefree polynomial over the domain of this valuation

        - ``include_steps`` -- boolean (default: ``False``); where to include
          the valuations produced during the process of checking whether ``G``
          is totally ramified in the return value

        - ``assume_squarefree`` -- boolean (default: ``False``); whether to
          assume that ``G`` is square-free over the completion of the domain of
          this valuation. Setting this to ``True`` can significantly improve
          the performance.

        ALGORITHM:

        This is a simplified version of :meth:`sage.rings.valuation.valuation.DiscreteValuation.mac_lane_approximants`.

        EXAMPLES::

            sage: k = Qp(5,4)
            sage: v = k.valuation()
            sage: R.<x> = k[]
            sage: G = x^2 + 1
            sage: v.is_totally_ramified(G)                                              # needs sage.geometry.polyhedron
            False
            sage: G = x + 1
            sage: v.is_totally_ramified(G)
            True
            sage: G = x^2 + 2
            sage: v.is_totally_ramified(G)
            False
            sage: G = x^2 + 5
            sage: v.is_totally_ramified(G)                                              # needs sage.geometry.polyhedron
            True
            sage: v.is_totally_ramified(G, include_steps=True)                          # needs sage.geometry.polyhedron
            (True, [Gauss valuation induced by 5-adic valuation, [ Gauss valuation induced by 5-adic valuation, v((1 + O(5^4))*x) = 1/2 ]])

        We consider an extension as totally ramified if its ramification index
        matches the degree. Hence, a trivial extension is totally ramified::

            sage: R.<x> = QQ[]
            sage: v = QQ.valuation(2)
            sage: v.is_totally_ramified(x)
            True

        TESTS:

        An example that Sebastian Pauli used at Sage Days 87::

            sage: R = ZpFM(3, 20)
            sage: S.<x> = R[]
            sage: f = x^9 + 9*x^2 + 3
            sage: R.valuation().is_totally_ramified(f)                                  # needs sage.geometry.polyhedron
            True
        """
        R = G.parent()

        from sage.rings.polynomial.polynomial_ring import PolynomialRing_generic
        if not isinstance(R, PolynomialRing_generic) or R.base_ring() is not self.domain() or not G.is_monic():
            raise ValueError("G must be a monic univariate polynomial over the domain of this valuation")
        if not assume_squarefree and not G.is_squarefree():
            raise ValueError("G must be squarefree")

        from sage.rings.valuation.gauss_valuation import GaussValuation

        steps = [ GaussValuation(R, self) ]
        while True:
            v = steps[-1]
            if v.F() > 1:
                ret = False
                break
            if v.E() == G.degree():
                ret = True
                break

            assert v(G) is not infinity
            if v.is_key(G):
                ret = False
                break

            next = v.mac_lane_step(G, assume_squarefree=True)
            if len(next) > 1:
                ret = False
                break
            steps.append(next[0])

        if include_steps:
            return ret, steps
        return ret

    def change_domain(self, ring):
        r"""
        Change the domain of this valuation to ``ring`` if possible.

        EXAMPLES::

            sage: v = ZZ.valuation(2)
            sage: v.change_domain(QQ).domain()
            Rational Field
        """
        return pAdicValuation(ring, self.p())

    def _extensions_to_quotient(self, ring, approximants=None):
        r"""
        Return the extensions of this valuation to an integral quotient over
        the domain of this valuation.

        EXAMPLES::

            sage: R.<x> = QQ[]
            sage: QQ.valuation(2)._extensions_to_quotient(R.quo(x^2 + x + 1))
            [2-adic valuation]
        """
        approximants = approximants or self.mac_lane_approximants(ring.modulus().change_ring(self.domain()), assume_squarefree=True, require_incomparability=True)
        return [pAdicValuation(ring, approximant, approximants) for approximant in approximants]

    def extensions(self, ring):
        r"""
        Return the extensions of this valuation to ``ring``.

        EXAMPLES::

            sage: v = ZZ.valuation(2)
            sage: v.extensions(GaussianIntegers())                                      # needs sage.rings.number_field
            [2-adic valuation]

        TESTS::

            sage: R.<a> = QQ[]
            sage: x = polygen(ZZ, 'x')
            sage: L.<a> = QQ.extension(x^3 - 2)
            sage: R.<b> = L[]
            sage: M.<b> = L.extension(b^2 + 2*b + a)
            sage: M.valuation(2)
            2-adic valuation

        Check that we can extend to a field written as a quotient::

            sage: R.<x> = QQ[]
            sage: K.<a> = QQ.extension(x^2 + 1)
            sage: R.<y> = K[]
            sage: L.<b> = R.quo(x^2 + a)
            sage: QQ.valuation(2).extensions(L)
            [2-adic valuation]

        A case where there was at some point an internal error in the
        approximants code::

            sage: R.<x> = QQ[]
            sage: L.<a> = NumberField(x^4 + 2*x^3 + 2*x^2 + 8)
            sage: QQ.valuation(2).extensions(L)
            [[ 2-adic valuation, v(x + 2) = 3/2 ]-adic valuation,
             [ 2-adic valuation, v(x) = 1/2 ]-adic valuation]

        A case where the extension was incorrect at some point::

            sage: v = QQ.valuation(2)
            sage: L.<a> = NumberField(x^2 + 2)
            sage: M.<b> = L.extension(x^2 + 1)
            sage: w = v.extension(L).extension(M)
            sage: w(w.uniformizer())
            1/4

        A case where the extensions could not be separated at some point::

            sage: v = QQ.valuation(2)
            sage: R.<x> = QQ[]
            sage: F = x^48 + 120*x^45 + 56*x^42 + 108*x^36 + 32*x^33 + 40*x^30 + 48*x^27 + 80*x^24 + 112*x^21 + 96*x^18 + 96*x^15 + 24*x^12 + 96*x^9 + 16*x^6 + 96*x^3 + 68
            sage: L.<a> = QQ.extension(F)
            sage: v.extensions(L)
            [[ 2-adic valuation, v(x) = 1/24, v(x^24 + 4*x^18 + 10*x^12 + 12*x^6 + 8*x^3 + 6) = 29/8 ]-adic valuation,
             [ 2-adic valuation, v(x) = 1/24, v(x^24 + 4*x^18 + 2*x^12 + 12*x^6 + 8*x^3 + 6) = 29/8 ]-adic valuation]
        """
        if self.domain() is ring:
            return [self]
        domain_fraction_field = _fraction_field(self.domain())
        if domain_fraction_field is not self.domain():
            if domain_fraction_field.is_subring(ring):
                return pAdicValuation(domain_fraction_field, self).extensions(ring)
        if self.domain().is_subring(ring):
            from sage.rings.polynomial.polynomial_quotient_ring import PolynomialQuotientRing_generic
            if isinstance(ring, PolynomialQuotientRing_generic):
                if isinstance(self.domain(), PolynomialQuotientRing_generic):
                    if self.domain().modulus() == ring.modulus():
                        base_extensions = self._base_valuation.extensions(self._base_valuation.domain().change_ring(self._base_valuation.domain().base_ring().fraction_field()))
                        return [pAdicValuation(ring, base._initial_approximation) for base in base_extensions]
                if ring.base_ring() is self.domain():
                    from sage.categories.integral_domains import IntegralDomains
                    if ring in IntegralDomains():
                        return self._extensions_to_quotient(ring)
                elif self.domain().is_subring(ring.base_ring()):
                    return sum([w.extensions(ring) for w in self.extensions(ring.base_ring())], [])
            from sage.rings.number_field.number_field_base import NumberField
            if isinstance(ring.fraction_field(), NumberField):
                if ring.base_ring().fraction_field() is self.domain().fraction_field():
                    approximants = self.mac_lane_approximants(ring.fraction_field().relative_polynomial().change_ring(self.domain()), assume_squarefree=True, require_incomparability=True)
                    return [pAdicValuation(ring, approximant, approximants) for approximant in approximants]
                if ring.base_ring() is not ring and self.domain().is_subring(ring.base_ring()):
                    return sum([w.extensions(ring) for w in self.extensions(ring.base_ring())], [])
        return super().extensions(ring)

    def restriction(self, ring):
        r"""
        Return the restriction of this valuation to ``ring``.

        EXAMPLES::

            sage: v = GaussianIntegers().valuation(2)                                   # needs sage.rings.number_field
            sage: v.restriction(ZZ)                                                     # needs sage.rings.number_field
            2-adic valuation
        """
        if ring is self.domain():
            return self

        if not ring.is_subring(self.domain()):
            raise ValueError("ring must be a subring of the domain of this valuation but %r is not a subring of %r" % (ring, self.domain()))

        return pAdicValuation(ring, self.p())

    @cached_method
    def value_semigroup(self):
        r"""
        Return the value semigroup of this valuation.

        EXAMPLES::

            sage: v = GaussianIntegers().valuation(2)                                   # needs sage.rings.number_field
            sage: v.value_semigroup()                                                   # needs sage.rings.number_field
            Additive Abelian Semigroup generated by 1/2
        """
        from sage.categories.fields import Fields
        v = self(self.uniformizer())
        if self.domain() in Fields():
            return DiscreteValueSemigroup([-v,v])
        return DiscreteValueSemigroup([v])


class pAdicValuation_padic(pAdicValuation_base):
    """
    The `p`-adic valuation of a complete `p`-adic ring.

    INPUT:

    - ``R`` -- a `p`-adic ring

    EXAMPLES::

        sage: v = Qp(2).valuation(); v  # indirect doctest
        2-adic valuation

    TESTS::

        sage: TestSuite(v).run()                # long time                             # needs sage.geometry.polyhedron
    """
    def __init__(self, parent):
        """
        TESTS::

            sage: from sage.rings.padics.padic_valuation import pAdicValuation_padic
            sage: isinstance(Qp(2).valuation(), pAdicValuation_padic)
            True
        """
        pAdicValuation_base.__init__(self, parent, parent.domain().prime())

    def reduce(self, x):
        """
        Reduce ``x`` modulo the ideal of elements of positive valuation.

        INPUT:

        - ``x`` -- an element of the domain of this valuation

        OUTPUT: an element of the :meth:`~sage.rings.valuation.valuation_space.DiscretePseudoValuationSpace.ElementMethods.residue_field`

        EXAMPLES::

            sage: R = Zp(3)
            sage: Zp(3).valuation().reduce(R(4))
            1
        """
        x = self.domain().coerce(x)
        return self.residue_field()(x.residue())

    def lift(self, x):
        """
        Lift ``x`` from the :meth:`~sage.rings.valuation.valuation_space.DiscretePseudoValuationSpace.ElementMethods.residue_field` to the domain of this
        valuation.

        INPUT:

        - ``x`` -- an element of the residue field of this valuation

        EXAMPLES::

            sage: R = Zp(3)
            sage: v = R.valuation()
            sage: xbar = v.reduce(R(4))
            sage: v.lift(xbar)
            1 + O(3^20)
        """
        x = self.residue_field().coerce(x)
        return self.domain()(x).lift_to_precision()

    def uniformizer(self):
        """
        Return a uniformizer of this valuation.

        EXAMPLES::

            sage: v = Zp(3).valuation()
            sage: v.uniformizer()
            3 + O(3^21)
        """
        return self.domain().uniformizer()

    def element_with_valuation(self, v):
        """
        Return an element of valuation ``v``.

        INPUT:

        - ``v`` -- an element of the :meth:`pAdicValuation_base.value_semigroup` of this valuation

        EXAMPLES::

            sage: R = Zp(3)
            sage: v = R.valuation()
            sage: v.element_with_valuation(3)
            3^3 + O(3^23)

            sage: K = Qp(3)
            sage: R.<y> = K[]
            sage: L.<y> = K.extension(y^2 + 3*y + 3)
            sage: L.valuation().element_with_valuation(3/2)
            y^3 + O(y^43)
        """
        from sage.rings.integer_ring import ZZ
        from sage.rings.rational_field import QQ
        v = QQ(v)
        if v not in self.value_semigroup():
            raise ValueError("%r is not in the value semigroup of %r" % (v, self))
        v = ZZ(v * self.domain().absolute_e())
        return self.domain().one() << v

    def _repr_(self):
        """
        Return a printable representation of this valuation.

        EXAMPLES::

            sage: ZZ.valuation(3)._repr_()
            '3-adic valuation'
        """
        return "%s-adic valuation" % (self.p())

    def _call_(self, x):
        r"""
        Evaluate this valuation at ``x``.

        EXAMPLES::

            sage: K = Qp(3)
            sage: R.<y> = K[]                                                           # needs sage.libs.ntl
            sage: L.<y> = K.extension(y^2 - 3)                                          # needs sage.libs.ntl
            sage: L.valuation()(3)                                                      # needs sage.libs.ntl
            1
        """
        return x.ordp()

    def residue_ring(self):
        r"""
        Return the residue field of this valuation.

        EXAMPLES::

            sage: Qq(9, names='a').valuation().residue_ring()                           # needs sage.libs.ntl
            Finite Field in a0 of size 3^2
        """
        return self.domain().residue_field()

    def shift(self, x, s):
        r"""
        Shift ``x`` in its expansion with respect to :meth:`uniformizer` by
        ``s`` "digits".

        For nonnegative ``s``, this just returns ``x`` multiplied by a
        power of the uniformizer `\pi`.

        For negative ``s``, it does the same but when not over a field, it
        drops coefficients in the `\pi`-adic expansion which have negative
        valuation.

        EXAMPLES::

            sage: R = ZpCA(2)
            sage: v = R.valuation()
            sage: v.shift(R.one(), 1)
            2 + O(2^20)
            sage: v.shift(R.one(), -1)
            O(2^19)

            sage: S.<y> = R[]
            sage: S.<y> = R.extension(y^3 - 2)
            sage: v = S.valuation()
            sage: v.shift(1, 5)
            y^5 + O(y^60)
        """
        x = self.domain().coerce(x)
        s = self.value_group()(s)
        return x << s

    def simplify(self, x, error=None, force=False):
        r"""
        Return a simplified version of ``x``.

        Produce an element which differs from ``x`` by an element of
        valuation strictly greater than the valuation of ``x`` (or strictly
        greater than ``error`` if set.)

        INPUT:

        - ``x`` -- an element in the domain of this valuation

        - ``error`` -- a rational, infinity, or ``None`` (default: ``None``),
          the error allowed to introduce through the simplification

        - ``force`` -- ignored

        EXAMPLES::

            sage: R = Zp(2)
            sage: v = R.valuation()
            sage: v.simplify(6)
            2 + O(2^21)
            sage: v.simplify(6, error=0)
            0
        """
        x = self.domain().coerce(x)

        if error is None:
            error = self(x)
        from sage.rings.infinity import infinity
        if error is infinity:
            return x
        # we need to scale by the ramification index because p-adics use a
        # different normalization
        normalized_error = (error / self.value_group().gen()).ceil()
        return x.add_bigoh(normalized_error + 1).lift_to_precision()


class pAdicValuation_int(pAdicValuation_base):
    r"""
    A `p`-adic valuation on the integers or the rationals.

    EXAMPLES::

        sage: v = ZZ.valuation(3); v
        3-adic valuation

    TESTS::

        sage: TestSuite(v).run()                # long time                             # needs sage.geometry.polyhedron
    """
    def _repr_(self):
        """
        Return a printable representation of this valuation.

        EXAMPLES::

            sage: ZZ.valuation(3)._repr_()
            '3-adic valuation'
        """
        return "%s-adic valuation" % (self.p())

    def _call_(self, x):
        """
        Evaluate this valuation at ``x``.

        INPUT:

        - ``x`` -- an element in the domain of this valuation

        EXAMPLES::

            sage: ZZ.valuation(3)(9)
            2
        """
        if x.is_zero():
            # x.valuation() is a factor 10 slower when computing the valuation
            # of a rational zero than when computing the valuation of another
            # small rational. Special casing this is a factor 100 faster.
            return infinity
        return x.valuation(self._p)

    def uniformizer(self):
        """
        Return a uniformizer of this `p`-adic valuation, i.e., `p` as an
        element of the domain.

        EXAMPLES::

            sage: v = ZZ.valuation(3)
            sage: v.uniformizer()
            3
        """
        return self.domain()(self.p())

    def residue_ring(self):
        """
        Return the residue field of this valuation.

        EXAMPLES::

            sage: v = ZZ.valuation(3)
            sage: v.residue_ring()
            Finite Field of size 3
        """
        from sage.rings.finite_rings.finite_field_constructor import GF
        return GF(self.p())

    def _ge_(self, other):
        r"""
        Return whether this valuation is greater than or equal than ``other``
        everywhere.

        EXAMPLES::

            sage: v = ZZ.valuation(2)
            sage: w = valuations.TrivialValuation(ZZ)
            sage: v >= w
            True
        """
        if other.is_trivial():
            return other.is_discrete_valuation()
        if isinstance(other, pAdicValuation_int):
            return self.p() == other.p()
        return super(pAdicValuation_base, self)._ge_(other)

    def _relative_size(self, x):
        r"""
        Return an estimate on the coefficient size of ``x``.

        The number returned is an estimate on the factor between the number of
        bits used by ``x`` and the minimal number of bits used by an element
        congruent to ``x``.

        This is used by :meth:`simplify` to decide whether simplification of
        coefficients is going to lead to a significant shrinking of the
        coefficients of ``x``.

        EXAMPLES::

            sage: v = ZZ.valuation(2)
            sage: v._relative_size(2)
            1
            sage: v._relative_size(2**20)
            11
        """
        x = self.domain().coerce(x)
        return (x.numerator().nbits() + x.denominator().nbits())//self.p().nbits()

    def simplify(self, x, error=None, force=False, size_heuristic_bound=32):
        r"""
        Return a simplified version of ``x``.

        Produce an element which differs from ``x`` by an element of
        valuation strictly greater than the valuation of ``x`` (or strictly
        greater than ``error`` if set.)

        INPUT:

        - ``x`` -- an element in the domain of this valuation

        - ``error`` -- a rational, infinity, or ``None`` (default: ``None``),
          the error allowed to introduce through the simplification

        - ``force`` -- ignored

        - ``size_heuristic_bound`` -- when ``force`` is not set, the expected
          factor by which the ``x`` need to shrink to perform an actual
          simplification (default: 32)

        EXAMPLES::

            sage: v = ZZ.valuation(2)
            sage: v.simplify(6, force=True)
            2
            sage: v.simplify(6, error=0, force=True)
            0

        In this example, the usual rational reconstruction misses a good answer
        for some moduli (because the absolute value of the numerator is not
        bounded by the square root of the modulus)::

            sage: v = QQ.valuation(2)
            sage: v.simplify(110406, error=16, force=True)
            562/19
            sage: Qp(2, 16)(110406).rational_reconstruction()
            Traceback (most recent call last):
            ...
            ArithmeticError: rational reconstruction of 55203 (mod 65536) does not exist
        """
        if not force and self._relative_size(x) <= size_heuristic_bound:
            return x

        x = self.domain().coerce(x)

        v = self(x)
        if error is None:
            error = v
        from sage.rings.infinity import infinity
        if error is infinity:
            return x
        if error < v:
            return self.domain().zero()

        from sage.rings.rational_field import QQ
        from sage.rings.padics.factory import Qp
        precision_ring = Qp(self.p(), QQ(error).floor() + 1 - v)
        reduced = precision_ring(x)
        lift = (reduced >> v).lift()
        best = self.domain()(lift) * self.p()**v

        if self._relative_size(x) < self._relative_size(best):
            best = x

        # We implement a modified version of the usual rational reconstruction
        # algorithm (based on the extended Euclidean algorithm) here. We do not
        # get the uniqueness properties but we do not need them actually.
        # This is certainly slower than the implementation in Cython.
        from sage.categories.fields import Fields
        m = self.p()**(QQ(error).floor() + 1 - v)
        if self.domain() in Fields():
            r = (m, lift)
            s = (0, 1)
            while r[1]:
                qq, rr = r[0].quo_rem(r[1])
                r = r[1], rr
                s = s[1], s[0] - qq*s[1]
                from sage.arith.misc import GCD as gcd
                if s[1] != 0 and gcd(s[1], r[1]) == 1:
                    rational = self.domain()(r[1]) / self.domain()(s[1]) * self.p()**v
                    if self._relative_size(rational) < self._relative_size(best):
                        best = rational

        assert (self(x-best) > error)

        return best

    def inverse(self, x, precision):
        r"""
        Return an approximate inverse of ``x``.

        The element returned is such that the product differs from 1 by an
        element of valuation at least ``precision``.

        INPUT:

        - ``x`` -- an element in the domain of this valuation

        - ``precision`` -- a rational or infinity

        EXAMPLES::

            sage: v = ZZ.valuation(2)
            sage: x = 3
            sage: y = v.inverse(3, 2); y
            3
            sage: x*y - 1
            8

        This might not be possible for elements of positive valuation::

            sage: v.inverse(2, 2)
            Traceback (most recent call last):
            ...
            ValueError: element has no approximate inverse in this ring

        Unless the precision is very small::

            sage: v.inverse(2, 0)
            1
        """
        if not x.is_zero():
            y = ~x
            if y in self.domain():
                return self.domain()(y)
        if precision <= 0:
            return self.domain().one()

        from sage.rings.infinity import infinity
        if self(x) > 0 or precision is infinity:
            raise ValueError("element has no approximate inverse in this ring")

        from sage.rings.integer_ring import ZZ
        from sage.rings.rational_field import QQ
        return self.domain()(ZZ(x).inverse_mod(self.p() ** QQ(precision).ceil()))


class pAdicFromLimitValuation(FiniteExtensionFromLimitValuation, pAdicValuation_base):
    r"""
    A `p`-adic valuation on a number field or a subring thereof, i.e., a
    valuation that extends the `p`-adic valuation on the integers.

    EXAMPLES::

        sage: v = GaussianIntegers().valuation(3); v                                    # needs sage.rings.number_field
        3-adic valuation

    TESTS::

        sage: TestSuite(v).run(skip='_test_shift')      # long time                     # needs sage.rings.number_field

    The ``_test_shift`` test fails because the parent of the shift is
    incorrect, see :issue:`23971`::

        sage: v.shift(1, -1).parent()                                                   # needs sage.rings.number_field
        Number Field in I with defining polynomial x^2 + 1 with I = 1*I
    """
    def __init__(self, parent, approximant, G, approximants):
        r"""
        TESTS::

            sage: v = GaussianIntegers().valuation(3)                                   # needs sage.rings.number_field
            sage: from sage.rings.padics.padic_valuation import pAdicFromLimitValuation
            sage: isinstance(v, pAdicFromLimitValuation)                                # needs sage.rings.number_field
            True
        """
        FiniteExtensionFromLimitValuation.__init__(self, parent, approximant, G, approximants)
        pAdicValuation_base.__init__(self, parent, approximant.restriction(approximant.domain().base_ring()).p())

    def _to_base_domain(self, f):
        r"""
        Return ``f``, an element of the underlying limit valuation, as an
        element of the domain of this valuation.

        EXAMPLES::

            sage: v = GaussianIntegers().valuation(3)                                   # needs sage.rings.number_field
            sage: I = GaussianIntegers().fraction_field().gen()                         # needs sage.rings.number_field
            sage: v._to_base_domain(I)                                                  # needs sage.rings.number_field
            x

        TESTS:

        Check that this also works for relative extensions::

            sage: v = QQ.valuation(2)
            sage: x = polygen(ZZ, 'x')
            sage: L.<a> = NumberField(x^2 + 2)
            sage: M.<b> = L.extension(x^2 + 1)
            sage: w = v.extension(L).extension(M)
            sage: w._to_base_domain(b)
            x
        """
        polynomial = f.lift()
        return polynomial(self._base_valuation.domain().gen())

    def _from_base_domain(self, f):
        r"""
        Return ``f``, an element of the domain of this valuation, as an element
        of the domain of the underlying limit valuation.

        EXAMPLES::

            sage: v = GaussianIntegers().valuation(3)                                   # needs sage.rings.number_field
            sage: v._from_base_domain(v._base_valuation.domain().gen())                 # needs sage.rings.number_field
            I
        """
        return self.domain()(f)

    def extensions(self, ring):
        r"""
        Return the extensions of this valuation to ``ring``.

        EXAMPLES::

            sage: v = GaussianIntegers().valuation(3)                                   # needs sage.rings.number_field
            sage: v.extensions(v.domain().fraction_field())                             # needs sage.rings.number_field
            [3-adic valuation]
        """
        if ring is self.domain().fraction_field():
            if self.domain() is not self.domain().fraction_field():
                G = ring.relative_polynomial()
                approximant = self._base_valuation.change_domain(G.parent())._initial_approximation
                return [pAdicValuation(ring, approximant)]
        return super().extensions(ring)


def _fraction_field(ring):
    r"""
    Return a fraction field of ``ring``.

    EXAMPLES:

    This works around some annoyances with ``ring.fraction_field()``::

        sage: R.<x> = ZZ[]
        sage: S = R.quo(x^2 + 1)
        sage: S.fraction_field()
        Fraction Field of Univariate Quotient Polynomial Ring in xbar over Integer Ring with modulus x^2 + 1

        sage: from sage.rings.padics.padic_valuation import _fraction_field
        sage: _fraction_field(S)
        Univariate Quotient Polynomial Ring in xbar over Rational Field with modulus x^2 + 1
    """
    from sage.categories.fields import Fields
    if ring in Fields():
        return ring

    from sage.rings.polynomial.polynomial_quotient_ring import PolynomialQuotientRing_generic
    if isinstance(ring, PolynomialQuotientRing_generic):
        from sage.categories.integral_domains import IntegralDomains
        if ring in IntegralDomains():
            return ring.base().change_ring(ring.base_ring().fraction_field()).quo(ring.modulus())
    return ring.fraction_field()



from sage.libs.eclib.constructor import CremonaModularSymbols
from sage.libs.eclib.interface import mwrank_EllipticCurve, mwrank_MordellWeil
from sage.libs.eclib.mwrank import get_precision, set_precision
from sage.libs.eclib.mwrank import initprimes as mwrank_initprimes
